# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2021- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr

__all__ = ['fftn', 'ifftn', 'rfftn', 'irfftn', 'dctn', 'idctn', 'dstn', 'idstn',
           'vkfft_version', 'clear_vkfftapp_cache',
           'has_pycuda', 'has_opencl', 'has_cupy']

from enum import Enum
from functools import lru_cache
import numpy as np
from .base import complex32, strides_nonzero
from . import config

try:
    from .cuda import VkFFTApp as VkFFTApp_cuda, has_pycuda, has_cupy, vkfft_version

    if has_pycuda:
        import pycuda.gpuarray as cua
        import pycuda.driver as cu_drv
    if has_cupy:
        import cupy as cp
except (ImportError, OSError):
    has_cupy, has_pycuda = False, False

try:
    from .opencl import VkFFTApp as VkFFTApp_cl, cla, vkfft_version

    has_opencl = True
except (ImportError, OSError):
    has_opencl = False


class Backend(Enum):
    """ Backend language & library"""
    UNKNOWN = 0
    PYCUDA = 1
    PYOPENCL = 2
    CUPY = 3


def _prepare_transform(src, dest, cl_queue, cuda_stream, r2c=False, r2c_odd=False):
    """
    Determine the backend from the input data.
    Create the destination array if necessary.

    :param src: the source GPU array
    :param dest: the destination array. If None, a new GPU array is created.
    :param cl_queue: the opencl queue to use, or None
    :param cuda_stream: the cuda stream to use (from cupy or pycuda), or None
    :param r2c: if True, this is for an R2C transform, so adapt the destination
        array accordingly.
    :param r2c_odd: True if the r2c transform has an odd size (for the real
        array) along the x-axis. This is only needed for a c2r out-of-place
        transform to determine the destination x-axis size.
    :return: a tuple (backend, inplace, dest, cl_queue, cuda_stream, devctx),
        also appending the destination dtype for a r2c transform.
        devctx is either the device or context unique ptr or id - this
        is only used to cache the VkFFTapp (e.g. making sure the app is
        re-instantiated if the cuda device changes). The cuda_stream
        output will be the int ptr.
    """
    backend = Backend.UNKNOWN
    # fast axis is the last only for C-ordered arrays
    # We don't use np.argmin(src.strides) because of special cases
    # with axis size=1
    fast_axis = src.ndim - 1 - np.argmin(strides_nonzero(src.strides)[::-1])
    if fast_axis == src.ndim - 1:
        order = 'C'
    else:
        order = 'F'
    if r2c:
        if src.dtype in [np.float16, np.float32, np.float64]:
            sh = list(src.shape)
            sh[fast_axis] = sh[fast_axis] // 2 + 1
            dtype = np.complex64
            if src.dtype == np.float16:
                dtype = complex32
            elif src.dtype == np.float64:
                dtype = np.complex128
        else:
            sh = list(src.shape)
            sh[fast_axis] = (sh[fast_axis] - 1) * 2
            if r2c_odd:
                # This is only true for an out-of-place transform,
                # but sh is only used if dest is None
                sh[fast_axis] += 1
            dtype = np.float32
            if src.dtype == complex32:
                dtype = np.float16
            elif src.dtype == np.complex128:
                dtype = np.float64
    else:
        sh, dtype = None, None
    devctx = None
    if has_pycuda:
        if isinstance(src, cua.GPUArray):
            backend = Backend.PYCUDA
            # Must cast the gpudata to int as it can either be a DeviceAllocation object
            # or an int (e.g. when using a view of another array)
            src_ptr = int(src.gpudata)
            if dest is None:
                if r2c:
                    dest = cua.empty(tuple(sh), dtype=dtype, allocator=src.allocator, order=order)
                else:
                    dest = cua.empty_like(src)
            dest_ptr = int(dest.gpudata)
            if cuda_stream is None:
                devctx = cu_drv.Context.get_current()
            elif isinstance(cuda_stream, cu_drv.Stream):
                # Pass an int to make sure it is hashable
                cuda_stream = cuda_stream.handle

    if backend == Backend.UNKNOWN and has_opencl:
        if isinstance(src, cla.Array):
            backend = Backend.PYOPENCL
            src_ptr = src.data.int_ptr
            if dest is None:
                if r2c:
                    dest = cla.empty(src.queue, tuple(sh), dtype=dtype, allocator=src.allocator, order=order)
                else:
                    dest = cla.empty_like(src)
            dest_ptr = dest.data.int_ptr
            if cl_queue is None:
                cl_queue = src.queue
            devctx = cl_queue.context

    if backend == Backend.UNKNOWN and has_cupy:
        if isinstance(src, cp.ndarray):
            backend = Backend.CUPY
            src_ptr = src.__cuda_array_interface__['data'][0]
            if dest is None:
                if r2c:
                    dest = cp.empty(tuple(sh), dtype=dtype, order=order)
                else:
                    dest = cp.empty_like(src)
            dest_ptr = dest.__cuda_array_interface__['data'][0]
            if cuda_stream is None:
                cuda_stream = cp.cuda.get_current_stream()
            if isinstance(cuda_stream, cp.cuda.Stream):
                cuda_stream = cuda_stream.ptr
            devctx = cp.cuda.Device().id

    if backend == Backend.UNKNOWN:
        raise RuntimeError("Could note determine the type of GPU array supplied, or the "
                           "corresponding backend is not installed "
                           "(has_pycuda=%d, has_pyopencl=%d, has_cupy=%d)" %
                           (has_pycuda, has_opencl, has_cupy))

    inplace = dest_ptr == src_ptr
    if r2c:
        if inplace:
            dest = src.view(dtype=dtype)
        return backend, inplace, dest, cl_queue, cuda_stream, devctx, dtype
    else:
        return backend, inplace, dest, cl_queue, cuda_stream, devctx


@lru_cache(maxsize=config.FFT_CACHE_NB)
def _get_fft_app(backend, shape, dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue,
                 devctx, strides=None, tune=False):
    del devctx  # Variable is just used for proper lru_cache
    sback = {Backend.PYCUDA: 'pycuda', Backend.CUPY: 'cupy', Backend.PYOPENCL: 'pyopencl'}[backend]
    tune_config = {'backend': sback} if tune else None
    if backend in [Backend.PYCUDA, Backend.CUPY]:
        return VkFFTApp_cuda(shape, dtype, ndim=ndim, inplace=inplace,
                             stream=cuda_stream, norm=norm, axes=axes, strides=strides,
                             tune_config=tune_config)
    elif backend == Backend.PYOPENCL:
        return VkFFTApp_cl(shape, dtype, cl_queue, ndim=ndim, inplace=inplace,
                           norm=norm, axes=axes, strides=strides,
                           tune_config=tune_config)


@lru_cache(maxsize=config.FFT_CACHE_NB)
def _get_rfft_app(backend, shape, dtype, inplace, ndim, norm, cuda_stream, cl_queue,
                  devctx, strides=None, tune=False, r2c_odd=False):
    del devctx  # Variable is just used for proper lru_cache
    sback = {Backend.PYCUDA: 'pycuda', Backend.CUPY: 'cupy', Backend.PYOPENCL: 'pyopencl'}[backend]
    tune_config = {'backend': sback} if tune else None
    if backend in [Backend.PYCUDA, Backend.CUPY]:
        return VkFFTApp_cuda(shape, dtype, ndim=ndim, inplace=inplace,
                             stream=cuda_stream, norm=norm, r2c=True, strides=strides,
                             tune_config=tune_config, r2c_odd=r2c_odd)
    elif backend == Backend.PYOPENCL:
        return VkFFTApp_cl(shape, dtype, cl_queue, ndim=ndim, inplace=inplace,
                           norm=norm, r2c=True, strides=strides,
                           tune_config=tune_config, r2c_odd=r2c_odd)


@lru_cache(maxsize=config.FFT_CACHE_NB)
def _get_dct_app(backend, shape, dtype, inplace, ndim, norm, dct_type,
                 cuda_stream, cl_queue, devctx, tune=False):
    del devctx  # Variable is just used for proper lru_cache
    sback = {Backend.PYCUDA: 'pycuda', Backend.CUPY: 'cupy', Backend.PYOPENCL: 'pyopencl'}[backend]
    tune_config = {'backend': sback} if tune else None
    if backend in [Backend.PYCUDA, Backend.CUPY]:
        return VkFFTApp_cuda(shape, dtype, ndim=ndim, inplace=inplace,
                             stream=cuda_stream, norm=norm, dct=dct_type,
                             tune_config=tune_config)
    elif backend == Backend.PYOPENCL:
        return VkFFTApp_cl(shape, dtype, cl_queue, ndim=ndim, inplace=inplace,
                           norm=norm, dct=dct_type,
                           tune_config=tune_config)


@lru_cache(maxsize=config.FFT_CACHE_NB)
def _get_dst_app(backend, shape, dtype, inplace, ndim, norm, dst_type,
                 cuda_stream, cl_queue, devctx, tune=False):
    del devctx  # Variable is just used for proper lru_cache
    sback = {Backend.PYCUDA: 'pycuda', Backend.CUPY: 'cupy', Backend.PYOPENCL: 'pyopencl'}[backend]
    tune_config = {'backend': sback} if tune else None
    if backend in [Backend.PYCUDA, Backend.CUPY]:
        return VkFFTApp_cuda(shape, dtype, ndim=ndim, inplace=inplace,
                             stream=cuda_stream, norm=norm, dst=dst_type,
                             tune_config=tune_config)
    elif backend == Backend.PYOPENCL:
        return VkFFTApp_cl(shape, dtype, cl_queue, ndim=ndim, inplace=inplace,
                           norm=norm, dst=dst_type,
                           tune_config=tune_config)


def fftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_queue=None,
         return_scale=False, tune=False):
    """
    Perform a FFT on a GPU array, automatically creating the VkFFTApp
    and caching it for future re-use.

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: if 0 (un-normalised), every transform multiplies the L2 norm
        of the array by the transform size.
        if 1 (the default) or "backward", the inverse transform divides the
        L2 norm by the array size, so FFT+iFFT will keep the array norm.
        if "ortho", each transform will keep the L2 norm, but that will
        involve an extra read & write operation.
    :param axes: a list or tuple of axes along which the transform is made.
        if None, the transform is done along the ndim fastest axes, or all
        axes if ndim is None. Not allowed for R2C transforms
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param return_scale: if True, return the scale factor by which the result
        must be multiplied to keep its L2 norm after the transform
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :return: the destination array if return_scale is False, or (dest, scale)
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
    app = _get_fft_app(backend, src.shape, src.dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue, devctx,
                       strides=src.strides, tune=tune)
    if backend == Backend.PYOPENCL:
        app.fft(src, dest, queue=cl_queue)
    else:
        app.fft(src, dest)
    if return_scale:
        s = app.get_fft_scale()
        return dest, s
    return dest


def ifftn(src, dest=None, ndim=None, norm=1, axes=None, cuda_stream=None, cl_queue=None,
          return_scale=False, tune=False):
    """
    Perform an inverse FFT on a GPU array, automatically creating the VkFFTApp
    and caching it for future re-use.

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: if 0 (un-normalised), every transform multiplies the L2 norm
        of the array by the transform size.
        if 1 (the default) or "backward", the inverse transform divides the
        L2 norm by the array size, so FFT+iFFT will keep the array norm.
        if "ortho", each transform will keep the L2 norm, but that will
        involve an extra read & write operation.
    :param axes: a list or tuple of axes along which the transform is made.
        if None, the transform is done along the ndim fastest axes, or all
        axes if ndim is None. Not allowed for R2C transforms
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param return_scale: if True, return the scale factor by which the result
        must be multiplied to keep its L2 norm after the transform
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :return: the destination array if return_scale is False, or (dest, scale)
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
    app = _get_fft_app(backend, src.shape, src.dtype, inplace, ndim, axes, norm, cuda_stream, cl_queue, devctx,
                       strides=src.strides, tune=tune)
    if backend == Backend.PYOPENCL:
        app.ifft(src, dest, queue=cl_queue)
    else:
        app.ifft(src, dest)
    if return_scale:
        s = app.get_fft_scale()
        return dest, s
    return dest


def rfftn(src, dest=None, ndim=None, norm=1, cuda_stream=None, cl_queue=None,
          return_scale=False, tune=False, r2c_odd=False):
    """
    Perform a real->complex transform on a GPU array, automatically creating
    the VkFFTApp and caching it for future re-use.
    For an out-of-place transform, the length of the destination last axis will
    be src.shape[-1]//2+1.
    For an in-place transform with an even [respectively odd]-sized
    fast (x) axis, the src array should have a shape (..., nx+2)
    [respectively (..., nx+1)], the last one or two values along the
    fast (x) axis are ignored, and the destination
    array will have a shape of (..., nx//2+1).
    An in-place transform with an odd-sized x-axis requires r2c_odd=True

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: if 0 (un-normalised), every transform multiplies the L2 norm
        of the array by the transform size.
        if 1 (the default) or "backward", the inverse transform divides the
        L2 norm by the array size, so FFT+iFFT will keep the array norm.
        if "ortho", each transform will keep the L2 norm, but that will
        involve an extra read & write operation.
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param return_scale: if True, return the scale factor by which the result
        must be multiplied to keep its L2 norm after the transform
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :param r2c_odd: should be set to True for an in-place r2c transform
        where the actual data length is odd along the fast axis. This
        parameter is ignored otherwise.
    :return: the destination array if return_scale is False, or (dest, scale).
        For an in-place transform, the returned value is a view of the array
        with the appropriate type.
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx, dtype = \
        _prepare_transform(src, dest, cl_queue, cuda_stream, True)
    app = _get_rfft_app(backend, src.shape, src.dtype, inplace, ndim, norm, cuda_stream, cl_queue, devctx,
                        strides=src.strides, tune=tune, r2c_odd=r2c_odd)
    if backend == Backend.PYOPENCL:
        app.fft(src, dest, queue=cl_queue)
    else:
        app.fft(src, dest)
    if return_scale:
        s = app.get_fft_scale()
        return dest.view(dtype=dtype), s
    return dest.view(dtype=dtype)


def irfftn(src, dest=None, ndim=None, norm=1, cuda_stream=None, cl_queue=None,
           return_scale=False, tune=False, r2c_odd=False):
    """
    Perform a complex->real transform on a GPU array, automatically creating
    the VkFFTApp and caching it for future re-use.
    For an out-of-place transform, the length of the destination last axis will
    be (src.shape[-1]-1)*2.
    For an in-place transform, if the src complex array has a shape (..., nx),
    the destination (real) array will have a shape of (..., nx*2), but the last
    one (if r2c_odd=True) or two values along the x-axis are used as buffer:
    the size of the transform is thus either nx*2 or nx*2+1.

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: if 0 (un-normalised), every transform multiplies the L2 norm
        of the array by the transform size.
        if 1 (the default) or "backward", the inverse transform divides the
        L2 norm by the array size, so FFT+iFFT will keep the array norm.
        if "ortho", each transform will keep the L2 norm, but that will
        involve an extra read & write operation.
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param return_scale: if True, return the scale factor by which the result
        must be multiplied to keep its L2 norm after the transform
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :param r2c_odd: should be set to True for an in-place r2c transform
        where the actual data length (in the real array) is odd along the
        fast axis. This parameter is ignored otherwise.
    :return: the destination array if return_scale is False, or (dest, scale)
        For an in-place transform, the returned value is a view of the array
        with the appropriate type.
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx, dtype = \
        _prepare_transform(src, dest, cl_queue, cuda_stream, True, r2c_odd=r2c_odd)
    app = _get_rfft_app(backend, dest.shape, dest.dtype, inplace, ndim, norm, cuda_stream, cl_queue, devctx,
                        strides=dest.strides, tune=tune, r2c_odd=r2c_odd)
    if backend == Backend.PYOPENCL:
        app.ifft(src, dest, queue=cl_queue)
    else:
        app.ifft(src, dest)
    if return_scale:
        s = app.get_fft_scale()
        return dest.view(dtype=dtype), s
    return dest.view(dtype=dtype)


def dctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_queue=None, tune=False):
    """
    Perform a real->real Direct Cosine Transform on a GPU array, automatically
    creating the VkFFTApp and caching it for future re-use.

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: normalisation mode, either 0 (un-normalised) or
        1 (the default, also available as "backward) which will normalise
        the inverse transform, so DCT+iDCT will keep the array norm.
    :param dct_type: the type of dct desired: 1, 2 (default), 3 or 4
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :return: the destination array.
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
    app = _get_dct_app(backend, src.shape, src.dtype, inplace, ndim, norm,
                       dct_type, cuda_stream, cl_queue, devctx, tune=tune)
    if backend == Backend.PYOPENCL:
        app.fft(src, dest, queue=cl_queue)
    else:
        app.fft(src, dest)
    return dest


def idctn(src, dest=None, ndim=None, norm=1, dct_type=2, cuda_stream=None, cl_queue=None, tune=False):
    """
    Perform a real->real inverse Direct Cosine Transform on a GPU array,
    automatically creating the VkFFTApp and caching it for future re-use.

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: normalisation mode, either 0 (un-normalised) or
        1 (the default, also available as "backward) which will normalise
        the inverse transform, so DCT+iDCT will keep the array norm.
    :param dct_type: the type of dct desired: 2 (default), 3 or 4
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :return: the destination array.
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
    app = _get_dct_app(backend, src.shape, src.dtype, inplace, ndim, norm,
                       dct_type, cuda_stream, cl_queue, devctx, tune=tune)
    if backend == Backend.PYOPENCL:
        app.ifft(src, dest, queue=cl_queue)
    else:
        app.ifft(src, dest)
    return dest


def dstn(src, dest=None, ndim=None, norm=1, dst_type=2, cuda_stream=None, cl_queue=None, tune=False):
    """
    Perform a real->real Direct Cosine Transform on a GPU array, automatically
    creating the VkFFTApp and caching it for future re-use.

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: normalisation mode, either 0 (un-normalised) or
        1 (the default, also available as "backward) which will normalise
        the inverse transform, so DST+iDST will keep the array norm.
    :param dst_type: the type of dst desired: 1, 2 (default), 3 or 4
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :return: the destination array.
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
    app = _get_dst_app(backend, src.shape, src.dtype, inplace, ndim, norm,
                       dst_type, cuda_stream, cl_queue, devctx, tune=tune)
    if backend == Backend.PYOPENCL:
        app.fft(src, dest, queue=cl_queue)
    else:
        app.fft(src, dest)
    return dest


def idstn(src, dest=None, ndim=None, norm=1, dst_type=2, cuda_stream=None, cl_queue=None, tune=False):
    """
    Perform a real->real inverse Direct Cosine Transform on a GPU array,
    automatically creating the VkFFTApp and caching it for future re-use.

    :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
    :param dest: the destination GPU array. If None, a new GPU array will
        be created and returned (using the source array allocator
        (pycuda, pyopencl) if available).
        If dest is the same array as src, an inplace transform is done.
    :param ndim: the number of dimensions (<=3) to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param norm: normalisation mode, either 0 (un-normalised) or
        1 (the default, also available as "backward) which will normalise
        the inverse transform, so DST+iDST will keep the array norm.
    :param dst_type: the type of dst desired: 2 (default), 3 or 4
    :param cuda_stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. If None, the default one will be used
    :param cl_queue: the pyopencl.CommandQueue to be used. If None,
        the source array default queue will be used
    :param tune: if True, will activate the automatic tuning of VkFFT
        parameters to maximise the FT throughput. This uses a quick
        approach testing a few transforms (about 4) before choosing the
        optimal parameters. This is similar to FFTW's FFTW_MEASURE approach.
    :return: the destination array.
    """
    backend, inplace, dest, cl_queue, cuda_stream, devctx = _prepare_transform(src, dest, cl_queue, cuda_stream, False)
    app = _get_dst_app(backend, src.shape, src.dtype, inplace, ndim, norm,
                       dst_type, cuda_stream, cl_queue, devctx, tune=tune)
    if backend == Backend.PYOPENCL:
        app.ifft(src, dest, queue=cl_queue)
    else:
        app.ifft(src, dest)
    return dest


def clear_vkfftapp_cache():
    """ Remove all cached VkFFTApp"""
    _get_fft_app.cache_clear()
    _get_rfft_app.cache_clear()



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2022- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr
#
#
# pyvkfft script to run short or long unit tests

import argparse
import os.path
import sys
import unittest
import time
import timeit
import socket
import psutil
import numpy as np
from pyvkfft.test import TestFFT, TestFFTSystematic
from pyvkfft.version import __version__, vkfft_version, vkfft_git_version


def cuda_info_str():
    try:
        # The functions used should not create a context on the GPU
        from ..cuda import cuda_compile_version, cuda_runtime_version, cuda_driver_version
        return "CUDA: driver %s, runtime %s, compiled %s" % \
            (cuda_driver_version(), cuda_runtime_version(), cuda_compile_version())
    except:
        return ""


def suite_default():
    suite = unittest.TestSuite()
    load_tests = unittest.defaultTestLoader.loadTestsFromTestCase
    suite.addTest(load_tests(TestFFT))
    return suite


def suite_systematic():
    suite = unittest.TestSuite()
    load_tests = unittest.defaultTestLoader.loadTestsFromTestCase
    suite.addTest(load_tests(TestFFTSystematic))
    return suite


def make_html_pre_post(overwrite=False):
    if ('pyvkfft-test1000.html' not in os.listdir()) or overwrite:
        # if this was called by a parent pyvkfft-test-suite, print that command
        com = ""
        for c in psutil.Process(os.getppid()).cmdline():
            if "/" in c and "pyvkfft-test-suite" in c:
                c = os.path.split(c)[-1]
            com += "%s " % c
        if 'pyvkfft-test-suite' in com:
            com = 'Command:<tt>%s</tt>\n' % com
        else:
            com = ''
        # Need the html header, styles and the results' table beginning
        vkfft_ver = f'{vkfft_version()}' if 'unknown' in vkfft_git_version() \
            else f'{vkfft_version()}[{vkfft_git_version()}]'
        tmp = '<!DOCTYPE html>\n <html>\n <head> <style>\n' \
              'th, td { border: 1px solid grey;}\n' \
              '.center {margin-left: auto;  margin-right: auto; text-align:center;' \
              ' width:100%%; max-width:1920px; margin-left: auto;  margin-right: auto}\n' \
              '.cell_transform {background-color: #ccf;}\n' \
              '.active, .cell_transform:hover {background-color: #aaf;}\n' \
              '.toggle_graph {' \
              '  background-color: transparent;' \
              '  border: none;' \
              '  cursor: pointer;' \
              '  padding:0;' \
              '  outline: none;' \
              '  height: 100%%' \
              '  width: 100%%' \
              '}\n' \
              '.cell_error {background-color: #fcc;}\n' \
              '.active, .cell_error:hover {background-color: #faa;}\n' \
              '.toggle_error {' \
              '  background-color: transparent;' \
              '  border: none;' \
              '  cursor: pointer;' \
              '  padding:0;' \
              '  outline: none;' \
              '  height: 100%%' \
              '  width: 100%%' \
              '}\n' \
              '.toggle_fail {' \
              '  background-color: transparent;' \
              '  border: none;' \
              '  cursor: pointer;' \
              '  padding:0;' \
              '  outline: none;' \
              '  height: 100%%' \
              '  width: 100%%' \
              '}\n' \
              '.label_ok {' \
              'background-color: #00ff00;' \
              'font-weight: bold;' \
              'color: #000;' \
              '}\n' \
              '</style>\n' \
              '</head>\n' \
              '<body>\n' \
              '<div class="center">' \
              '<h2>pyVkFFT test results</h2>\n' \
              '<h3>pyvkfft: %s,  VkFFFT:%s %s host : %s</h3>\n' \
              '<div style="text-align:left;">' \
              '<p>%s<p>Methodology: the included graphs measure the accuracy of the forward ' \
              'and backward transforms: an array is generated with random uniform values ' \
              'between -0.5 and 0.5, and the results of its transform are compared ' \
              'with either pyfftw (in long double precision) if available, or scipy if ' \
              'available, or numpy fft. The L2 curve measures the average square norm ' \
              'difference, and the L<sub>&infin;</sub> the maximum difference.' \
              '<p>Note: for the R2C inverse transform, the result of the forward ' \
              'transform is used instead of re-using the random array (in order to have ' \
              'a proper half-Hermitian array), contrary to what is done for other ' \
              'transforms. This explains why the IFFT R2C maximum (L<sub>&infin;</sub>) ' \
              'errors are larger.' \
              '<p>Note 2: some "errors" for DCT may be due to unsupported sizes in VkFFT, ' \
              'which vary depending on the card and language used (amount of ' \
              'shared/local memory). So they just indicate a current limit for the ' \
              'transform sizes rather than a real error.' \
              '<p> [Click on the highlighted cells for details and accuracy graphs ' \
              'vs the transform size]<br>\n' \
              '</div><table class="center">\n' \
              '   <thead>\n' \
              '       <tr>\n' \
              '           <th>GPU</th>' \
              '           <th>backend</th>' \
              '           <th>transform</th>' \
              '           <th>ndim</th>' \
              '           <th>range</th>' \
              '           <th>radix</th>' \
              '           <th>dtype</th>' \
              '           <th>inplace</th>' \
              '           <th>LUT</th>' \
              '           <th>norm</th>' \
              '           <th>time-duration</th>' \
              '           <th>FAIL</th>' \
              '           <th>ERROR</th>' \
              '       </tr>\n' \
              '   </thead>\n' \
              '<tbody class="center">\n' % (__version__, vkfft_ver, cuda_info_str(),
                                            socket.gethostname(), com)
        open("pyvkfft-test1000.html", "w").write(tmp)
    if ('pyvkfft-test1999.html' not in os.listdir()) or overwrite:
        tmp = '</tbody>\n' \
              '</tbody\n' \
              '</table>\n' \
              '</div>\n' \
              '  <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.5.1/jquery.min.js"></script>' \
              '     <script id="rendered-js" >\n' \
              '        $(document).ready(function ()' \
              '        {\n' \
              '           $(".toggle_graph").click(function () \n' \
              '              {\n' \
              '                 var trgraph = $(this).parents().nextUntil(".row_results", ".graph"); \n' \
              '                 trgraph.toggle(); \n' \
              '                 if(trgraph.is( ":visible" ))\n' \
              '                 { trgraph.children("td:first").html($(this).data("command")+"<br>' \
              '                      <img src=\\"" + $(this).data("img") + "\\">");}\n' \
              '                 else { trgraph.children("td:first").html("img hidden");};\n' \
              '                 \n' \
              '               });\n' \
              '           $(".toggle_fail").click(function () ' \
              '              {$(this).parents().nextUntil(".row_results", ".failures").toggle();  });\n' \
              '           $(".toggle_error").click(function () ' \
              '              {$(this).parents().nextUntil(".row_results", ".errors").toggle();  });\n' \
              '        });\n' \
              '</script>' \
              '' \
              '</body>\n' \
              '</html>'
        open("pyvkfft-test1999.html", "w").write(tmp)


def name_next_file(pattern="pyvkfft-test%04d.html"):
    """
    Find the first unused name for a file, starting at i=1
    :param pattern: the pattern for the file name.
    :return: the filename
    """
    lsdir = os.listdir()
    for i in range(1001, 1999):
        if pattern % i not in lsdir:
            return pattern % i
    raise RuntimeError("name_next_file: '%s' files all used from 1001 to 1998. Maybe cleanup ?" % pattern)


def make_parser():
    epilog = "Examples:\n" \
             "  pyvkfft-test\n" \
             "      the regular test which tries the fft interface, using parallel\n" \
             "      streams (for pycuda), and C2C/R2C/DCT/DST transforms for sizes N=15,17,30,34\n" \
             "      with 1D to 4 or 5D transforms, also N=808,2988,4200,13000,13001," \
             "      13002,130172 for 1D and 2D transforms.\n" \
             "      All tests are done with single and double precision, in and\n" \
             "      out-of-place, norm=0 and 1, and all available backends (pyopencl,\n" \
             "      pycuda and cupy). For C2C arrays up to dimension 5 are tested,\n" \
             "      with all possible combination of transform axes.\n" \
             "      That's for a total of a tens of thousands transforms, which are tested\n" \
             "      against the result of numpy, scipy or pyfftw (when available) for\n" \
             "      accuracy.\n" \
             "      The text output gives the N2 and Ninf (aka max) relative norm of\n" \
             "      the transform, with the ratio in () to the expected tolerance for\n" \
             "      both direct and inverse transforms.\n" \
             "\n" \
             "  pyvkfft-test --quick\n" \
             "      Much faster version of the regular test, should only take a few\n" \
             "      seconds instead of minutes or 10s of minutes.\n" \
             "\n" \
             "  pyvkfft-test --nproc 8 --gpu v100 --mailto_fail toto@pyvkfft.org\n" \
             "      same test, but using 8 parallel process to speed up, and use a GPU\n" \
             "      with 'v100' in its name. Also, send the results in case of a failure\n" \
             "      to the given email address\n" \
             "\n" \
             "  pyvkfft-test --systematic --backend pycuda --nproc 8 --radix --range 2 10000\n" \
             "      Perform a systematic test of C2C transforms in (by default) 1D and\n" \
             "      single precision, for N=2 to 10000, only for radix transforms\n" \
             "\n" \
             "  pyvkfft-test --systematic --backend pycuda --nproc 8 --radix 2 7 11 --range 2 10000 --double\n" \
             "      Same test, but only for radix sizes with factors 2, 7 and 11, and double accuracy\n" \
             "\n" \
             "  pyvkfft-test --systematic --backend cupy --nproc 8 --bluestein --range 2 10000 --ndim 2 " \
             "--lut --inplace\n" \
             "      test with cupy backend, only non-radix 2D inplace R2C transforms\n," \
             "      using a lookup table( lut) for higher single precision accuracy.\n\n\n" \
             "Columns in the text output:\n" \
             " * backend\n" \
             " * type of transform\n" \
             " * array shape\n" \
             " * axes for the transform. If None, axes are set by the number of transform dimensions\n" \
             " * number of dimensions for the transform. Can be None if axes are given.\n" \
             " * type of algorithm for each axis: r=radix, R=Rader, B=Bluestein, -=skipped axis\n" \
             " * number of uploads for each axis: 0 if not transformed, 1 if the axis length fits\n" \
             "   in the cache and the transform can be done in 1 read+write, 2 or 3 if multi-upload is used\n" \
             " * data type and precision\n" \
             " * use of a Look-Up-Table (LUT) -for single precision only.\n" \
             " * inplace or out-of-place transform\n" \
             " * normalisation for the transform: 0 or 1\n" \
             " * order of the array: C (fast axis is last) or F (fast axis is first)\n" \
             " * N2 and N_inf error norm for the forward transform, with the comparison \n" \
             "   to the maximum allowed error (and in parenthesis the ratio to this maximum), \n" \
             "   and finally 0 or 1 depending on whether the source array was modified (0) or not (1)\n" \
             " * Same values for the inverse transform\n" \
             " * temporary buffer size allocated by VkFFT if necessary, for large transforms\n" \
             " * status: OK, FAIL (if accuracy is above limit or source array unexpectedly changed)\n" \
             "   or ERROR (an error was raised during execution, e.g. compilation, memory,...)\n"
    parser = argparse.ArgumentParser(prog='pyvkfft-test', epilog=epilog,
                                     description='Run pyvkfft unit tests, regular or systematic',
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('--colour', action='store_true',
                        help="Use colour depending on how good the measured accuracy is")
    parser.add_argument('--html', action='store', nargs='*',
                        help="Summarises the results in html row(s). This is saved to "
                             "'pyvkfft-test%%04d.html', starting at i=1001 and incrementing. "
                             "Files with i=1000 and i=1999 are the beginning and the end of the"
                             "html file, which can be concatenated to form a valid html page."
                             "If --graph is also used, this includes a graph of the accuracy "
                             "which can be displayed by clicking on the type of transform.")
    parser.add_argument('--gpu', action='store',
                        help="Name (or sub-string) of the GPU to use")
    parser.add_argument('--opencl_platform', action='store',
                        help="Name (or sub-string) of the opencl platform to use (case-insensitive). "
                             "Note that by default the PoCL platform is skipped, "
                             "unless it is specifically requested or it is the only one available "
                             "(PoCL has some issues with VkFFT for some transforms)")
    parser.add_argument('--mailto', action='store',
                        help="Email address the results will be sent to")
    parser.add_argument('--mailto_fail', action='store',
                        help="Email address the results will be sent to, only if the test fails")
    parser.add_argument('--mailto_smtp', action='store', default="localhost",
                        help="SMTP server address to mail the results")
    parser.add_argument('--nproc', action='store', nargs=1,
                        help="Number of parallel process to use to speed up tests. "
                             "Make sure the sum of parallel process will not use too much "
                             "GPU memory",
                        default=[1], type=int)
    parser.add_argument('--silent', action='store_true',
                        help="Use this to minimise the written output "
                             "(note that tests can take a long time be patient")
    parser.add_argument('--c2c', action='store_true',
                        help="When used without --systematic, perform only "
                             "c2c quick tests and skip the long r2c/dct/dst "
                             "unless they were also requested.")
    parser.add_argument('--systematic', action='store_true',
                        help="Perform a systematic accuracy test over a range of array sizes.\n"
                             "Without this argument a faster test (a few minutes) will be "
                             "performed with selected array sizes for all possible transforms.")
    parser.add_argument('--quick', action='store_true',
                        help="Only run a small subset of tests (c2c and r2c), with only "
                             "1D and 2D inplace, no extra dimensions or skipped axes, "
                             "no F-ordering and only single precision.")
    sysgrp = parser.add_argument_group("systematic", "Options for --systematic:")
    sysgrp.add_argument('--axes', action='store', nargs='*', type=int,
                        help="transform axes: x (fastest) is 1,"
                             "y is 2, z is 3, e.g. '--axes 1', '--axes 2 3'."
                             "The default is to perform the transform along the ndim fastest "
                             "axes. Using this overrides --ndim")
    sysgrp.add_argument('--backend', action='store', nargs='+',
                        help="Choose single or multiple GPU backends,"
                             "by default all available backends are selected.",
                        choices=['pycuda', 'cupy', 'pyopencl'])
    sysgrp.add_argument('--bluestein', '--nonradix', action='store_true',
                        help="Only perform transform with non-radix dimensions, i.e. the "
                             "largest number in the prime decomposition of each array dimension "
                             "must be larger than 13")
    sysgrp.add_argument('--db', nargs='*', action='store',
                        help="Save the results to an sql database. If no filename is"
                             "given, pyvkfft-test.sql will be used. If the file already"
                             "exists, the results are added to the file. Fields stored"
                             "include HOSTNAME, EPOCH, BACKEND, LANGUAGE, TRANSFORM (c2c, r2c or "
                             "dct/dst1/2/3/4, AXES, ARRAY_SHAPE, NDIMS, NDIM, PRECISION, INPLACE,"
                             "NORM, LUT, N, N2_FFT, N2_IFFT, NI_FFT, NI_IFFT, TOLERANCE,"
                             "DT_APP, DT_FFT, DT_IFFT, SRC_UNCHANGED_FFT, SRC_UNCHANGED_IFFT, "
                             "GPU_NAME, SUCCESS, ERROR, VKFFT_ERROR_CODE")
    sysgrp.add_argument('--dct', nargs='?', action='store', type=int,
                        help="Test direct cosine transforms (default is c2c):"
                             " '--dct' (defaults to dct 2), '--dct 1'",
                        const=2, default=False, choices=[1, 2, 3, 4])
    sysgrp.add_argument('--dst', nargs='?', action='store', type=int,
                        help="Test direct sine transforms (default is c2c):"
                             " '--dst' (defaults to dst 2), '--dst 1'",
                        const=2, default=False, choices=[1, 2, 3, 4])
    sysgrp.add_argument('--double', action='store_true',
                        help="Use double precision (float64/complex128) instead of single")
    sysgrp.add_argument('--dry-run', action='store_true',
                        help="Perform a dry-run, printing the number of array shapes to test")
    parser.add_argument('--fast-random', action='store', default=None,
                        help="Use this option to run a random percentage of the full test suite, "
                             "for faster results. A number between 5 and 100 is required.", type=int)
    sysgrp.add_argument('--inplace', action='store_true',
                        help="Use inplace transforms")
    sysgrp.add_argument('--graph', action='store', nargs="?", const="",
                        help="Save the graph of the accuracy as a function of the size"
                             "to the given filename (if no name is given, it will be "
                             "automatically generated)."
                             "Requires matplotlib, and scipy for linear regression.")
    sysgrp.add_argument('--lut', action='store_true',
                        help="Force the use of a LUT for the transform, to improve accuracy. "
                             "By default VkFFT will activate the LUT on some GPU with less "
                             "accurate accelerated trigonometric functions. "
                             "This is automatically true for double precision")
    sysgrp.add_argument('--max-nb-tests', action='store', nargs=1,
                        help="Maximum number of tests. If the number of generated test "
                             "cases is larger, the program will abort.",
                        default=[1000], type=int)
    sysgrp.add_argument('--ndim', action='store', nargs=1,
                        help="Number of dimensions for the transform. Using 12 or 123 "
                             "will result in testing bother 1 and 2 or 1,2 and 3. It is"
                             "recommended to use --range_mb and ",
                        default=[1], type=int, choices=[1, 2, 3, 12, 123])
    # sysgrp.add_argument('--ndims', action='store', nargs=1,
    #                     help="Number of dimensions for the array (must be >=ndim). "
    #                          "By default, the array will have the same dimensionality "
    #                          "as the transform (ndim)",
    #                     type=int, choices=[1, 2, 3, 4])
    sysgrp.add_argument('--norm', action='store', nargs=1, type=int,
                        help="Normalisation to test (must be 1 for dct or dst)",
                        default=[1], choices=[0, 1])
    sysgrp.add_argument('--ref-long-double', action='store_true',
                        help="Use long double precision for the reference calculation, "
                             "(requires scipy). This gives more objective accuracy plots but "
                             "can be slower (or much slower on some architectures).")
    sysgrp.add_argument('--r2c', action='store_true', help="Test real-to-complex transform "
                                                           "(default is c2c)")
    sysgrp.add_argument('--fstride', action='store_true',
                        help="Test F-ordered arrays (default is C-ordered). Not supported for DCT/DST")
    sysgrp.add_argument('--radix', action='store', nargs='*', type=int,
                        help="Perform only radix transforms. If no value is given, all available "
                             "radix transforms are allowed. Alternatively a list can be given: "
                             "'--radix 2' (only 2**n array sizes), '--radix 2 3 5' "
                             "(only 2**N1 * 3**N2 * 5**N3)",
                        choices=[2, 3, 5, 7, 11, 13])
    sysgrp.add_argument('--radix-max-pow', action='store', nargs=1, type=int,
                        help="For radix runs, specify the maximum exponent of each base "
                             "integer, i.e. for '--radix 2 3 --radix-max-pow 2' will limit "
                             "lengths to 2**N1 * 3**N2 with N1,N2<=2")
    sysgrp.add_argument('--range', action='store', nargs=2, type=int,
                        help="Range of array lengths [min, max] along each transform dimension, "
                             "'--range 2 128'",
                        default=[2, 128])
    sysgrp.add_argument('--range-mb', action='store', nargs=2, type=int,
                        help="Allowed range of array sizes [min, max] in Mbytes, e.g. "
                             "'--range-mb 2 128'. This can be used to limit the arrays size "
                             "while allowing large lengths along individual dimensions. "
                             "It can also be used to separate runs with a given size range "
                             "and different nproc values. This takes into account the "
                             "type (single or double), and also whether the transform "
                             "is made inplace, so this represents the total GPU memory"
                             "used.",
                        default=[0, 128])
    sysgrp.add_argument('--range-nd-narrow', action='store', nargs=2, default=['0', '0'],
                        help="Two values (drel dabs), e.g. '--range_nd_narrow 0.10 11' "
                             "with 0<=drel<=1 and dabs (integer>=0) must be given "
                             "to allow 2D and 3D tests to be done on arrays with different "
                             "lengths along every dimension, but while limiting the "
                             "difference between lengths. For example in 2D for an "
                             "(N1,N2) array shape, generated lengths will verify "
                             "abs(n2-n1)<max(dabs+drel*N1). The default value of (0,0) "
                             "only allows the same lengths. This allows to test more "
                             "diverse configurations while limiting the number of tests.")
    sysgrp.add_argument('--serial', action='store_true',
                        help="Serialise the tests instead of spawning them in separate "
                             "process, allowing to diagnose more errors. Incompatible "
                             "with nproc>1.")
    sysgrp.add_argument('--timeout', action='store', nargs=1, type=int,
                        help="Change the timeout (in seconds) to raise a TimeOut error for "
                             "individual tests. After 4 have failed, give up.",
                        default=[120])
    return parser


def main():
    t0 = timeit.default_timer()
    localt0 = time.localtime()
    parser = make_parser()

    # parser.print_help()
    args = parser.parse_args()
    if args.serial and args.nproc[0] > 1:
        raise RuntimeError("Cannot use --serial with --nproc")
    if args.graph is not None:
        if not len(args.graph):
            args.graph = name_next_file("pyvkfft-test%03d.svg")

    # We modify class attributes to pass arguments - not a great approach but works..
    if args.systematic:
        t = TestFFTSystematic
        t.axes = args.axes
        t.bluestein = args.bluestein
        t.colour = args.colour
        t.dct = args.dct
        t.dst = args.dst
        t.db = args.db[0] if args.db is not None else None
        t.dry_run = args.dry_run
        t.dtype = np.float64 if args.double else np.float32
        t.fast_random = args.fast_random
        if args.fast_random is not None:
            assert 5 <= args.fast_random <= 100, "--fast-random must be between 5 and 100"
        t.gpu = args.gpu
        t.opencl_platform = args.opencl_platform
        t.graph = args.graph
        t.inplace = args.inplace
        t.lut = args.lut
        t.max_nb_tests = args.max_nb_tests[0]
        t.ndim = args.ndim[0]
        # t.ndims = args.ndims
        t.norm = args.norm[0]
        t.nproc = args.nproc[0]
        t.opencl_platform = args.opencl_platform
        t.r2c = args.r2c
        t.fstride = args.fstride
        t.radix = args.radix
        t.ref_long_double = args.ref_long_double
        t.max_pow = None if args.radix_max_pow is None else args.radix_max_pow[0]
        t.range = args.range
        size_min_max = np.array(args.range_mb) * 1024 ** 2 // 8
        if args.r2c or args.dct or args.dst:
            size_min_max = size_min_max * 2
        if args.double:
            size_min_max = size_min_max / 2
        if args.inplace:
            size_min_max = size_min_max / 2
        t.range_size = size_min_max.tolist()
        t.range_nd_narrow = float(args.range_nd_narrow[0]), int(args.range_nd_narrow[1])
        t.serial = args.serial
        t.timeout = args.timeout[0]
        t.vbackend = args.backend
        t.verbose = not args.silent
        t.vn = args.range
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(t)
        if t.verbose:
            res = unittest.TextTestRunner(verbosity=2).run(suite)
        else:
            res = unittest.TextTestRunner(verbosity=1).run(suite)
        if t.dry_run:
            print(t.nb_shapes_gen)
            sys.exit()
    else:
        t = TestFFT
        t.verbose = not args.silent
        t.colour = args.colour
        t.gpu = args.gpu
        t.nproc = args.nproc[0]
        t.opencl_platform = args.opencl_platform
        t.vbackend = args.backend
        t.quick = args.quick
        if args.backend is not None:
            # Remove tests depending on backend used
            test_cuda_stream = False
            for b in args.backend:
                if 'pycuda' in b.lower():
                    test_cuda_stream = True
            if not test_cuda_stream:
                del t.test_pycuda_streams
            if 'pyopencl' not in args.backend:
                del t.test_pyopencl_queues
        if args.r2c or args.dct or args.dst or args.c2c or args.quick:
            # A selection of subtests was made - remove the other long ones
            if not args.r2c and not (args.quick or args.c2c):
                del t.test_r2c
            if not args.dct:
                del t.test_dct
            if not args.dst:
                del t.test_dst
            if not args.c2c and not (args.quick or args.r2c):
                del t.test_c2c
            del t.test_squeeze

        print("Starting the regular test")
        if not t.quick:
            print("WARNING: "
                  "this can take a long time (10s of minutes) to test "
                  "many configurations.\n"
                  "You can use instead 'pyvkfft-test --quick' if you just "
                  "want to make a more basic functionality test.\n")
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(t)
        if t.verbose:
            res = unittest.TextTestRunner(verbosity=2).run(suite)
        else:
            res = unittest.TextTestRunner(verbosity=1).run(suite)

    sub = os.path.split(sys.argv[0])[-1]
    for i in range(1, len(sys.argv)):
        arg = sys.argv[i]
        if 'mail' not in arg and 'mail' not in sys.argv[i - 1] and 'html' not in arg and 'graph' not in arg:
            sub += " " + arg
    info = "Ran:\n   %s\n\n Result:%s\n\n" % (sub, "OK" if res.wasSuccessful() else "FAILURE")

    info += "Elapsed time for tests: %s\n\n" % time.strftime("%Hh %Mm %Ss", time.gmtime(timeit.default_timer() - t0))

    nb_err_fail = len(res.errors) + len(res.failures)

    if args.html is not None:
        make_html_pre_post(overwrite=False)
        html = ''
        # One row for the summary
        html += '<tr class="row_results">'
        html += '<td>%s</td><td>' % (args.gpu if args.gpu is not None else '-')
        if args.backend is not None:
            for a in args.backend:
                html += a + ' '
        else:
            html += 'all'
        html += '</td>'
        if args.systematic:
            has_graph = False
            if args.graph is not None:
                if os.path.exists(args.graph):
                    has_graph = True
            if has_graph:
                tmp = '<td class="cell_transform"><input class="toggle_graph" type="button"' \
                      'data-img="%s" data-command="%s"' % (args.graph, sub)
                tmp += 'value="%s" style="width:100%%; height:100%%"></td>'
            else:
                tmp = '<td>%s</td>'
            if args.r2c:
                html += tmp % 'R2C'
            elif t.dct or t.dst:
                html += tmp % ('DCT%d' % t.dct) if t.dct else tmp % ('DST%d' % t.dst)
            else:
                html += tmp % 'C2C'
            if t.ndim == 12:
                html += "<td>1,2</td>"
            elif t.ndim == 123:
                html += "<td>1,2,3</td>"
            else:
                html += "<td>%d</td>" % t.ndim
            html += "<td>%d-%d" % (args.range[0], args.range[1])
            if (t.range_nd_narrow[0] > 0 or t.range_nd_narrow[1] > 0) and t.ndim > 1:
                html += " <small>[|N<sub>i</sub>-N<sub>1</sub>|<={%d; %d%%N<sub>1</sub>}]</small>" % \
                        (t.range_nd_narrow[1], int(t.range_nd_narrow[0] * 100))
            elif t.ndim > 1:
                html += ' (' + 'N,' * (t.ndim - 1) + 'N)'
            html += "</td>"
            if args.bluestein:
                html += "<td>Bluestein/Rader</td>"
            elif args.radix is None:
                html += "<td>-</td>"
            else:
                html += "<td>"
                for i in (args.radix if len(args.radix) else [2, 3, 5, 7, 11, 13]):
                    html += "%d," % i
                html = html[:-1]
                if t.max_pow is not None:
                    html += '<small>[^N,N<=%d]</small>' % t.max_pow
                html += '</td>'
            html += "<td>%s</td>" % ('float64' if args.double else 'float32')
            html += "<td>%s</td>" % ('inplace' if args.inplace else 'out-of-place')
            html += "<td>%s</td>" % ('True' if args.lut else 'Auto')
            html += "<td>%d</td>" % args.norm[0]
            nbts = '[%5d tests]' % t.nb_test
        else:
            html += ''
            html += '<td colspan="8">Regular multi-dimensional C2C/R2C/R2R test</td>'
            nbts = ''

        html += '<td>%s +%s %s</td>' % (time.strftime("%Y-%m-%d %Hh%M:%S", localt0),
                                        time.strftime("%Hh %Mm %Ss", time.gmtime(timeit.default_timer() - t0)), nbts)

        if len(res.failures):
            html += '<td class="cell_error"> <input class="toggle_fail" type="button" ' \
                    'value="%d" style="width:100%%; height:100%%"></td>' % len(res.failures)
        else:
            html += '<td class="label_ok"> 0 </td>'

        if len(res.errors):
            html += '<td class="cell_error"> <input class="toggle_error" type="button" ' \
                    'value="%d" style="width:100%%; height:100%%"></td>' % len(res.errors)
        else:
            html += '<td class="label_ok"> 0 </td>'
        html += '</tr>\n'

        if args.systematic and args.graph is not None:
            if os.path.exists(args.graph):
                # Do not put the img tag of the file, else it gets pre-loaded and
                # this can crash the browser when aggregating many tests.
                # It will only be added when clicking on the transform cell.
                html += '<tr class="graph" style="display: none"><td colspan=13></td></tr>'

    if len(res.errors):
        tmp = "%s\n\nERRORS:\n\n" % sub
        for t, s in res.errors:
            tid = t.id()
            tid1 = tid.split('(')[0].split('.')[-1]
            tid0, tid2 = tid.split('.' + tid1)
            tmp += "=" * 70 + "\n" + '%s %s [%s]:\n' % (tid1, tid2, tid0)
            tmp += "-" * 70 + "\n" + s + "\n"
        info += tmp
        if args.html is not None:
            html += '<tr class="errors" style="display: none; text-align: left; font-family: monospace">' \
                    '<td colspan=13><pre>%s</pre></td></tr>' % tmp

    if len(res.failures):
        tmp = "%s\n\nFAILURES:\n\n" % sub
        for t, s in res.failures:
            tid = t.id()
            tid1 = tid.split('(')[0].split('.')[-1]
            tid0, tid2 = tid.split('.' + tid1)
            tmp += "=" * 70 + "\n" + '%s %s [%s]:\n' % (tid1, tid2, tid0)
            tmp += "-" * 70 + "\n" + s + "\n"
        info += tmp
        if args.html is not None:
            html += '<tr class="failures" style="display: none; text-align: left; font-family: monospace">' \
                    '<td colspan=13><pre>%s</pre></td></tr>' % tmp

    if args.mailto_fail is not None and (nb_err_fail > 0) or args.mailto is not None:
        import smtplib
        try:
            from email.message import EmailMessage

            msg = EmailMessage()
            msg['From'] = '"pyvkfft" <%s>' % args.mailto if args.mailto is not None else args.mailto_fail
            msg['To'] = msg['From']
            msg['Subject'] = '[fail=%d error=%d] %s' % \
                             (len(res.failures), len(res.errors), sub)
            print("Mailing results:\nFrom: %s\nTo: \n %sSubject: %s" % (msg['to'], msg['to'], msg['Subject']))
            msg.set_content(info)

            s = smtplib.SMTP(args.mailto_smtp)
            s.send_message(msg)
            s.quit()
        except (ConnectionRefusedError, smtplib.SMTPConnectError):
            print("Could not connect to SMTP server (%s) to send email." % args.mailto_smtp)

    if args.html is not None:
        if len(args.html):
            html_out = args.html[0]
        else:
            html_out = name_next_file("pyvkfft-test%03d.html")
        with open(html_out, 'w') as f:
            f.write(html)
    sys.exit(int(nb_err_fail > 0))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2023- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr

from copy import deepcopy
import itertools
import timeit
import numpy as np


def tune_vkfft(tune, shape, dtype: type, ndim=None, inplace=True, stream=None, queue=None,
               norm=1, r2c=False, dct=False, dst=False, axes=None, strides=None,
               verbose=False, r2c_odd=False, **kwargs):
    """
    Automatically test different configurations for a VkFFTApp, returning
    the set of parameters which maximise the FT throughput.
    The three parameters which are recommended to optimise are aimThreads,
    warpSize and coalescedMemory. Usually tuning a single one should suffice,
    but the right one could depend on the backend and GPU brand.

    Note that the GPU context must have been initialised before calling
    this function.

    :param tune: dictionary including the backend used and the parameter
        values which will be tested.
        This is EXPERIMENTAL, as wrong parameters may lead to crashes.
        Note that this will allocate temporary GPU arrays, unless the arrays
        to used have been passed as parameters ('dest' and 'src').
        Examples:
         * tune={'backend':'cupy'} - minimal example, will automatically test a small
           set of parameters (4 to 10 tests). Recommended !
         * tune={'backend':'pycuda', 'warpSize':[8,16,32,64,128]}: this will test
           5 possible values for the warpSize.
         * tune={'backend':'pyopencl', 'aimThreads':[32,64,128,256]}: this will test
           5 possible values for the warpSize.
         * tune={'backend':'cupy', 'groupedBatch':[[-1,-1,-1],[8,8,8], [4,16,16}:
           this will test 3 possible values for groupedBatch. This one is more
           tricky to use.
         * tune={'backend':'cupy, 'warpSize':[8,16,32,64,128], 'src':a}: this
           will test 5 possible values for the warpSize, with a given source GPU
           array. This would only be valid for an inplace transform as no
           destination array is given.
    :param shape: the shape of the array to be transformed. The number
        of dimensions of the array can be larger than the FFT dimensions,
        but only for 1D and 2D transforms. 3D FFT transforms can only
        be done on 3D arrays.
    :param dtype: the numpy dtype of the source array (can be complex64 or complex128)
    :param ndim: the number of dimensions to use for the FFT. By default,
        uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
        array to perform a batched 3D FFT on all the layers. The FFT
        is always performed along the last axes if the array's number
        of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
        on the x and y axes for ndim=2.
    :param inplace: if True (the default), performs an inplace transform and
        the destination array should not be given in fft() and ifft().
    :param stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
        for the transform. This can also be the pointer/handle (int) to the
        cuda stream object. If None, the default stream will be used.
    :param queue: the pyopencl CommandQueue to use for the transform.
    :param norm: if 0 (unnormalised), every transform multiplies the L2
        norm of the array by its size (or the size of the transformed
        array if ndim<d.ndim).
        if 1 (the default) or "backward", the inverse transform divides
        the L2 norm by the array size, so FFT+iFFT will keep the array norm.
        if "ortho", each transform will keep the L2 norm, but that will
        involve an extra read & write operation.
    :param r2c: if True, will perform a real->complex transform, where the
        complex destination is a half-hermitian array.
        For an inplace transform, if the input data shape is (...,nx), the input
        float array should have a shape of (..., nx+2), the last two columns
        being ignored in the input data, and the resulting
        complex array (using pycuda's GPUArray.view(dtype=np.complex64) to
        reinterpret the type) will have a shape (..., nx//2 + 1).
        For an out-of-place transform, if the input (real) shape is (..., nx),
        the output (complex) shape should be (..., nx//2+1).
        Note that for C2R transforms with ndim>=2, the source (complex) array
        is modified.
    :param dct: used to perform a Direct Cosine Transform (DCT) aka a R2R transform.
        An integer can be given to specify the type of DCT (1, 2, 3 or 4).
        if dct=True, the DCT type 2 will be performed, following scipy's convention.
    :param dst: used to perform a Direct Cosine Transform (DST) aka a R2R transform.
        An integer can be given to specify the type of DST (1, 2, 3 or 4).
        if dst=True, the DST type 2 will be performed, following scipy's convention.
    :param axes: a list or tuple of axes along which the transform should be made.
        if None, the transform is done along the ndim fastest axes, or all
        axes if ndim is None. For R2C transforms, the fast axis must be
        transformed.
    :param strides: the array strides - needed if not C-ordered.
    :param verbose: if True, print speed for each configuration
    :param r2c_odd: this should be set to True to perform an inplace r2c/c2r
        transform with an odd-sized fast (x) axis.
        Explanation: to perform a 1D inplace transform of an array with 100
            elements, the input array should have a 100+2 size, resulting in
            a half-Hermitian array of size 51. If the input data has a size
            of 101, the input array should also be padded to 102 (101+1), and
            the resulting half-Hermitian array also has a size of 51. A
            flag is thus needed to differentiate the cases of 100+2 or 101+1.
    :param: extra parameters passed on to VkFFT

    :raises RuntimeError: if the optimisation fails
    :return: (kw, res) where kw are the optimal kwargs which can be passed to the
        VkFFTApp creation routine, and res is the full set of results for the
        different configurations tested.
    """
    try:
        # Import
        if tune['backend'] == 'cupy':
            import cupy as cua
            from cupy.cuda import Event
            from .cuda import VkFFTApp
        elif tune['backend'] == 'pycuda':
            import pycuda.gpuarray as cua
            from pycuda.driver import Event
            from .cuda import VkFFTApp
        else:
            from .opencl import VkFFTApp
            import pyopencl as cl
            import pyopencl.array as cla
        # GPU arrays
        if 'src' not in tune:
            if tune['backend'] == 'pyopencl':
                src = cla.to_device(queue, np.ones(shape, dtype=dtype))
            else:
                src = cua.ones(shape, dtype=dtype)
        else:
            src = tune['src']
        if inplace:
            dest = src
        else:
            if 'dest' not in tune:
                if 'src' in tune:
                    raise RuntimeError("VkFFT autotune: 'src' array is provided but not the destination, "
                                       "which is required for an out-of-place transform")
                if r2c:
                    # TODO: handle different strides, with fast axis != -1
                    shc = list(shape)
                    shc[-1] = shape[-1] // 2 + 1
                    shc = tuple(shc)
                    dtypec = np.complex64 if dtype == np.float32 else np.complex128
                    if tune['backend'] == 'pyopencl':
                        dest = cla.to_device(queue, np.ones(shc, dtype=dtypec))
                    else:
                        dest = cua.ones(shc, dtype=np.float32).astype(dtypec)
                else:
                    dest = src.copy()
            else:
                dest = tune['dest']

        # Parameters to test
        vk = []
        for k in tune.keys():
            if k in ['disableReorderFourStep', 'coalescedMemory', 'numSharedBanks',
                     'aimThreads', 'performBandwidthBoost', 'registerBoost',
                     'registerBoostNonPow2', 'registerBoost4Step', 'warpSize',
                     'groupedBatch']:
                vk.append(k)

        if len(vk) == 0:
            # Only the backend was supplied, choose automatically the parameters to tune
            tune = deepcopy(tune)
            if tune['backend'] in ['cupy', 'pycuda']:
                vk.append('coalescedMemory')
                tune['coalescedMemory'] = [32, 64, 128]
            else:
                # pyopencl - choose tuning parameters based on platform
                if 'apple' in queue.device.name.lower():
                    vk.append('aimThreads')
                    tune['aimThreads'] = [32, 64, 128, 256]
                elif 'nvidia' in queue.device.platform.name.lower():
                    vk.append('coalescedMemory')
                    tune['coalescedMemory'] = [32, 64, 128]
                else:
                    # TODO: try other GPU/platforms
                    vk.append('aimThreads')
                    tune['aimThreads'] = [32, 64, 128, 256]

        if verbose:
            print('VkFFT parameters to tune:  ' + '  '.join(vk))
        res = []
        args = (tune[k] for k in vk)
        for v in itertools.product(*args):
            kw = deepcopy(kwargs)
            for i in range(len(v)):
                kw[vk[i]] = v[i]
            if tune['backend'] == 'pyopencl':
                app = VkFFTApp(shape, dtype=dtype, queue=queue, ndim=ndim, inplace=inplace,
                               norm=norm, r2c=r2c, dct=dct, dst=dst, axes=axes,
                               strides=strides, r2c_odd=r2c_odd, **kw)
            else:
                app = VkFFTApp(shape, dtype=dtype, ndim=ndim, inplace=inplace, stream=stream,
                               norm=norm, r2c=r2c, dct=dct, dst=dst, axes=axes,
                               strides=strides, r2c_odd=r2c_odd, **kw)
                start = Event()
                stop = Event()
            dt = 0
            # Repeat the transform enough to have a meaningful measurement
            niter = max(1, int(1e8 // src.nbytes))
            for i in range(3):  # Get best of 3 transform speed
                if tune['backend'] == 'pyopencl':
                    queue.finish()
                    t0 = timeit.default_timer()
                    for ii in range(niter):
                        dest = app.fft(src, dest)
                        src = app.ifft(dest, src)
                    queue.finish()
                    dt1 = timeit.default_timer() - t0
                else:
                    start.record()
                    for ii in range(niter):
                        dest = app.fft(src, dest)
                        src = app.ifft(dest, src)
                    stop.record()
                    stop.synchronize()
                    if tune['backend'] == 'cupy':
                        dt1 = cua.cuda.get_elapsed_time(start, stop) / 1000
                    else:
                        dt1 = stop.time_since(start) / 1000
                if dt == 0:
                    dt = dt1
                elif dt1 < dt:
                    dt = dt1

            gbps = niter * src.nbytes * app.ndim * 2 * 2 / dt / 1024 ** 3
            del app
            res.append((kw, gbps, dt))
            if verbose:
                s = f"VkFFT tune {shape}"
                for ii in range(len(v)):
                    s += f" {vk[ii]}={v[ii]}"
                s += f" dt={dt:.3f} {gbps:.3f} GB/s"
                print(s)
        res.sort(key=lambda x: x[1])
        return res[-1][0], res
    except Exception as ex:
        print(ex)
        import traceback
        print(traceback.format_exc())
        return kwargs, []



# -*- coding: utf-8 -*-

""" Global configuration variables. The approach is adapted
 from Numba's config.py
"""
import os

FFT_CACHE_NB = 32
"""
Number of VkFFTApp to cache through the pyvkfft.fft interface
This must be modified *before* importing pyvkfft.fft.
"""

USE_LUT = None
"""
Force using a LUT for single-precision transforms ?
If None, this will be activated automatically for some GPU (Intel).

Use only to improve the accuracy by a factor 3 or 4.

If useLUT is passed directly to a VkFFTApp, this is ignored

Valid values: either None or 1
"""

WARN_OPENCL_QUEUE_MISMATCH = True
"""
Emit warning it the queue is different in the application and the array
"""


def process_environ(environ):
    """Process environment variables and load defaults as needed"""
    if "PYVKFFT_FFT_CACHE_NB" in environ:
        FFT_CACHE_NB = eval(environ["PYVKFFT_FFT_CACHE_NB"])
    else:
        FFT_CACHE_NB = 32

    if "PYVKFFT_USE_LUT" in environ:
        USE_LUT = eval(environ["PYVKFFT_USE_LUT"])
    else:
        USE_LUT = None

    if "PYVKFFT_WARN_OPENCL_QUEUE_MISMATCH" in environ:
        WARN_OPENCL_QUEUE_MISMATCH = eval(environ["PYVKFFT_WARN_OPENCL_QUEUE_MISMATCH"])

    # Inject values into the module globals
    for name, value in locals().copy().items():
        if name.isupper():
            globals()[name] = value


class _EnvReloader(object):

    def __init__(self):
        self.reset()
        self.old_environ = {}

    def reset(self):
        self.old_environ = {}
        self.update(force=True)

    def update(self, force=False):
        new_environ = {}

        for name, value in os.environ.items():
            if name.startswith('PYVKFFT_'):
                print(name, value)
                new_environ[name] = value
        if force or self.old_environ != new_environ:
            process_environ(new_environ)
            self.old_environ = new_environ


_env_reloader = _EnvReloader()


def _reload_config():
    """
    Reload the configuration from environment variables, if necessary.
    """
    _env_reloader.update()



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2021- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr

import ctypes
import numpy as np
from .tune import tune_vkfft

try:
    import pycuda.driver as cu_drv

    has_pycuda = True
except ImportError:
    has_pycuda = False
try:
    import cupy as cp

    has_cupy = True
except ImportError:
    has_cupy = False
    import sys

    if has_pycuda is False and 'sphinx' not in sys.modules:
        raise ImportError("You need either PyCUDA or CuPy to use pyvkfft.cuda.")

from .base import load_library, VkFFTApp as VkFFTAppBase, check_vkfft_result, ctype_int_size_p

try:
    _vkfft_cuda = load_library("_vkfft_cuda")


    class _types:
        """Aliases"""
        vkfft_config = ctypes.c_void_p
        stream = ctypes.c_void_p
        vkfft_app = ctypes.c_void_p


    _vkfft_cuda.make_config.restype = ctypes.c_void_p
    _vkfft_cuda.make_config.argtypes = [ctype_int_size_p, ctypes.c_size_t,
                                        ctypes.c_void_p, ctypes.c_void_p, _types.stream,
                                        ctypes.c_int, ctypes.c_size_t, ctypes.c_int,
                                        ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                        ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_size_t,
                                        ctype_int_size_p, ctypes.c_int,
                                        ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                        ctypes.c_int, ctype_int_size_p, ctypes.c_int]

    _vkfft_cuda.init_app.restype = ctypes.c_void_p
    _vkfft_cuda.init_app.argtypes = [_types.vkfft_config, ctypes.POINTER(ctypes.c_int),
                                     ctypes.POINTER(ctypes.c_size_t),
                                     ctype_int_size_p, ctype_int_size_p]

    _vkfft_cuda.fft.restype = ctypes.c_int
    _vkfft_cuda.fft.argtypes = [_types.vkfft_app, ctypes.c_void_p, ctypes.c_void_p]

    _vkfft_cuda.ifft.restype = ctypes.c_int
    _vkfft_cuda.ifft.argtypes = [_types.vkfft_app, ctypes.c_void_p, ctypes.c_void_p]

    _vkfft_cuda.free_app.restype = None
    _vkfft_cuda.free_app.argtypes = [_types.vkfft_app]

    _vkfft_cuda.free_config.restype = None
    _vkfft_cuda.free_config.argtypes = [_types.vkfft_config]

    _vkfft_cuda.vkfft_max_fft_dimensions.restype = ctypes.c_uint32
    _vkfft_cuda.vkfft_max_fft_dimensions.argtypes = None
except OSError:
    # This is used for doc generation
    import sys

    if 'sphinx' in sys.modules:
        pass
    else:
        raise


class VkFFTApp(VkFFTAppBase):
    """
    VkFFT application interface, similar to a cuFFT plan.
    """

    def __init__(self, shape, dtype: type, ndim=None, inplace=True, stream=None, norm=1,
                 r2c=False, dct=False, dst=False, axes=None, strides=None, tune_config=None,
                 r2c_odd=False, verbose=False, **kwargs):
        """

        :param shape: the shape of the array to be transformed. The number
            of dimensions of the array can be larger than the FFT dimensions,
            but only for 1D and 2D transforms. 3D FFT transforms can only
            be done on 3D arrays.
        :param dtype: the numpy dtype of the source array (can be complex64 or complex128)
        :param ndim: the number of dimensions to use for the FFT. By default,
            uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
            array to perform a batched 3D FFT on all the layers. The FFT
            is always performed along the last axes if the array's number
            of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
            on the x and y axes for ndim=2.
        :param inplace: if True (the default), performs an inplace transform and
            the destination array should not be given in fft() and ifft().
        :param stream: the pycuda.driver.Stream or cupy.cuda.Stream to use
            for the transform. This can also be the pointer/handle (int) to the
            cuda stream object. If None, the default stream will be used.
        :param norm: if 0 (unnormalised), every transform multiplies the L2
            norm of the array by its size (or the size of the transformed
            array if ndim<d.ndim).
            if 1 (the default) or "backward", the inverse transform divides
            the L2 norm by the array size, so FFT+iFFT will keep the array norm.
            if "ortho", each transform will keep the L2 norm, but that will
            involve an extra read & write operation.
        :param r2c: if True, will perform a real->complex transform, where the
            complex destination is a half-hermitian array.
            For an inplace transform, if the input data shape is (...,nx), the input
            float array should have a shape of (..., nx+2) if nx is even
            or (..., nx+1) if nx is odd, the last one or two columns
            being ignored in the input data, and the resulting
            complex array (using pycuda's GPUArray.view(dtype=np.complex64) to
            reinterpret the type) will have a shape (..., nx//2 + 1).
            For an out-of-place transform, if the input (real) shape is (..., nx),
            the output (complex) shape should be (..., nx//2+1).
            Note that for C2R transforms with ndim>=2, the source (complex) array
            is modified.
            For an inplace transform with an odd-sized x-axis, see the r2c_odd
            parameter.
        :param dct: used to perform a Direct Cosine Transform (DCT) aka a R2R transform.
            An integer can be given to specify the type of DCT (1, 2, 3 or 4).
            if dct=True, the DCT type 2 will be performed, following scipy's convention.
        :param dst: used to perform a Direct Sine Transform (DST) aka a R2R transform.
            An integer can be given to specify the type of DST (1, 2, 3 or 4).
            if dst=True, the DST type 2 will be performed, following scipy's convention.
        :param axes: a list or tuple of axes along which the transform should be made.
            if None, the transform is done along the ndim fastest axes, or all
            axes if ndim is None. For R2C transforms, the fast axis must be
            transformed.
        :param strides: the array strides - needed if not C-ordered.
        :param tune_config: this can be used to automatically generate an
            optimised set of VkFFT parameters by testing various configurations
            and measuring the FFT speed, in a manner similar to fftw's FFTW_MEASURE.
            This should be a dictionary including the backend used and the parameter
            values which will be tested.
            This is EXPERIMENTAL, as wrong parameters may lead to crashes.
            Note that this will allocate temporary GPU arrays, unless the arrays
            to used have been passed as parameters ('dest' and 'src').
            Examples:
            tune={'backend':'cupy} - minimal example, will automatically test a small
            set of parameters (4 to 10 tests). Recommended !
            tune={'backend':'cupy, 'warpSize':[8,16,32,64,128]}: this will test
            5 possible values for the warpSize.
            tune={'backend':'cupy, 'groupedBatch':[[-1,-1,-1],[8,8,8], [4,16,16}:
            this will test 3 possible values for groupedBatch. This one is more
            tricky to use.
            tune={'backend':'cupy, 'warpSize':[8,16,32,64,128], 'src':a}: this
            will test 5 possible values for the warpSize, with a given source GPU
            array. This would only be valid for an inplace transform as no
            destination array is given.
        :param r2c_odd: this should be set to True to perform an inplace r2c/c2r
            transform with an odd-sized fast (x) axis.
            Explanation: to perform a 1D inplace transform of an array with 100
                elements, the input array should have a 100+2 size, resulting in
                a half-Hermitian array of size 51. If the input data has a size
                of 101, the input array should also be padded to 102 (101+1), and
                the resulting half-Hermitian array also has a size of 51. A
                flag is thus needed to differentiate the cases of 100+2 or 101+1.
        :param verbose: if True, print a 1-string info about this VkFFTApp.
            See __str__ for details.

        :raises RuntimeError: if the initialisation fails, e.g. if the CUDA
            driver has not been properly initialised, or if the transform dimensions
            are not allowed by VkFFT.
        """
        if tune_config is not None:
            kwargs = tune_vkfft(tune_config, shape=shape, dtype=dtype, ndim=ndim, inplace=inplace, stream=stream,
                                norm=norm, r2c=r2c, dct=dct, dst=dst, axes=axes, strides=strides, verbose=False,
                                r2c_odd=r2c_odd, **kwargs)[0]
        super().__init__(shape, dtype, ndim=ndim, inplace=inplace, norm=norm, r2c=r2c,
                         dct=dct, dst=dst, axes=axes, strides=strides, r2c_odd=r2c_odd, **kwargs)

        self.stream = stream

        self.config = self._make_config()
        if self.config is None:
            raise RuntimeError("Error creating VkFFTConfiguration. Was the CUDA context properly initialised ?")

        res = ctypes.c_int(0)
        # Size of tmp buffer allocated by VkFFT - if any
        tmp_buffer_nbytes = ctypes.c_size_t(0)
        # 0 or 1 for each axis, only if the Bluestein algorithm is used
        use_bluestein_fft = np.zeros(vkfft_max_fft_dimensions(), dtype='long')
        # number of axis upload per dimension
        num_axis_upload = np.zeros(vkfft_max_fft_dimensions(), dtype='long')

        self.app = _vkfft_cuda.init_app(self.config, ctypes.byref(res),
                                        ctypes.byref(tmp_buffer_nbytes),
                                        use_bluestein_fft, num_axis_upload)

        check_vkfft_result(res, shape, dtype, ndim, inplace, norm, r2c, dct, dst, axes, "cuda")

        if self.app is None:
            raise RuntimeError("Error creating VkFFTApplication. Was the CUDA driver initialised ?")
        if has_pycuda:
            # TODO: This is a kludge to keep a reference to the context, so that it is deleted
            #  after the app in __delete__, which throws an error if the context does not exist
            #  anymore. Except that we cannot be sure this is the right context, if a stream
            #  has been given because we don't have access to cuStreamGetCtx from python...
            self._ctx = cu_drv.Context.get_current()

        self.tmp_buffer_nbytes = np.int64(tmp_buffer_nbytes)
        self.use_bluestein_fft = [bool(n) for n in use_bluestein_fft[:len(self.shape)]]
        self.nb_axis_upload = [int(num_axis_upload[i] * (self.skip_axis[i] is False))
                               for i in range(len(self.shape))]

        if verbose:
            print(self)

    def __del__(self):
        """ Takes care of deleting allocated memory in the underlying
        VkFFTApplication and VkFFTConfiguration.
        """
        if self.app is not None:
            _vkfft_cuda.free_app(self.app)
        if self.config is not None:
            _vkfft_cuda.free_config(self.config)

    def _make_config(self):
        """ Create a vkfft configuration for a FFT transform"""
        if len(self.shape) > vkfft_max_fft_dimensions():
            raise RuntimeError(f"Too many FFT dimensions after collapsing non-transform axes: "
                               f"{len(self.shape)}>{vkfft_max_fft_dimensions()}")

        shape = np.ones(vkfft_max_fft_dimensions(), dtype='long')
        shape[:len(self.shape)] = self.shape

        skip = np.zeros(vkfft_max_fft_dimensions(), dtype='long')
        skip[:len(self.skip_axis)] = self.skip_axis

        grouped_batch = np.empty(vkfft_max_fft_dimensions(), dtype='long')
        grouped_batch.fill(-1)
        grouped_batch[:len(self.groupedBatch)] = self.groupedBatch

        if self.r2c and self.inplace:
            # the last one or two columns are ignored in the R array, and will be used
            # in the C array with a size nx//2+1
            if self.r2c_odd:
                shape[0] -= 1
            else:
                shape[0] -= 2

        s = 0
        if self.stream is not None:
            if has_pycuda:
                if isinstance(self.stream, cu_drv.Stream):
                    s = self.stream.handle
            if has_cupy:
                if isinstance(self.stream, cp.cuda.Stream):
                    s = self.stream.ptr
            if s == 0 and isinstance(self.stream, int):
                # Assume the ptr or handle was passed
                s = self.stream

        if self.norm == "ortho":
            norm = 0
        else:
            norm = self.norm

        # We pass fake buffer pointer addresses to VkFFT. The real ones will be
        # given when performing the actual FFT.
        dest_gpudata = 2
        if self.inplace:
            dest_gpudata = 0

        return _vkfft_cuda.make_config(shape, self.ndim, 1, dest_gpudata, s,
                                       norm, self.precision, int(self.r2c),
                                       int(self.dct), int(self.dst),
                                       int(self.disableReorderFourStep), int(self.registerBoost),
                                       int(self.use_lut), int(self.keepShaderCode),
                                       self.n_batch, skip,
                                       int(self.coalescedMemory), int(self.numSharedBanks),
                                       int(self.aimThreads), int(self.performBandwidthBoost),
                                       int(self.registerBoostNonPow2), int(self.registerBoost4Step),
                                       int(self.warpSize), grouped_batch,
                                       int(self.forceCallbackVersionRealTransforms))

    def fft(self, src, dest=None):
        """
        Compute the forward FFT

        :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
        :param dest: the destination GPU array. Should be None for an inplace transform
        :raises RuntimeError: in case of a GPU kernel launch error
        :return: the transformed array. For a R2C inplace transform, the complex view of the
            array is returned.
        """
        use_cupy = False
        if has_cupy:
            if isinstance(src, cp.ndarray):
                use_cupy = True
        if use_cupy:
            src_ptr = src.__cuda_array_interface__['data'][0]
        else:
            # Must cast the gpudata to int as it can either be a DeviceAllocation object
            # or an int (e.g. when using a view of another array)
            src_ptr = int(src.gpudata)
        if dest is not None:
            if use_cupy:
                dest_ptr = dest.__cuda_array_interface__['data'][0]
            else:
                dest_ptr = int(dest.gpudata)
        else:
            dest_ptr = src_ptr
        if self.inplace:
            if src_ptr != dest_ptr:
                raise RuntimeError("VkFFTApp.fft: dest is not None but this is an inplace transform")
            res = _vkfft_cuda.fft(self.app, int(src_ptr), int(src_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst, backend="cuda")
            if self.norm == "ortho":
                src *= self._get_fft_scale(norm=0)
            if self.r2c:
                if src.dtype == np.float32:
                    return src.view(dtype=np.complex64)
                elif src.dtype == np.float64:
                    return src.view(dtype=np.complex128)
            return src
        else:
            if dest is None:
                raise RuntimeError("VkFFTApp.fft: dest is None but this is an out-of-place transform")
            if src_ptr == dest_ptr:
                raise RuntimeError("VkFFTApp.fft: dest and src are identical but this is an out-of-place transform")
            if self.r2c:
                assert (dest.size == src.size // src.shape[self.fast_axis] * (src.shape[self.fast_axis] // 2 + 1))
            res = _vkfft_cuda.fft(self.app, int(src_ptr), int(dest_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst, backend="cuda")
            if self.norm == "ortho":
                dest *= self._get_fft_scale(norm=0)
            return dest

    def ifft(self, src, dest=None):
        """
        Compute the backward FFT

        :param src: the source pycuda.gpuarray.GPUArray or cupy.ndarray
        :param dest: the destination GPU array. Should be None for an inplace transform
        :raises RuntimeError: in case of a GPU kernel launch error
        :return: the transformed array. For a C2R inplace transform, the float view of the
            array is returned.
        """
        use_cupy = False
        if has_cupy:
            if isinstance(src, cp.ndarray):
                use_cupy = True
        if use_cupy:
            src_ptr = src.__cuda_array_interface__['data'][0]
        else:
            # Must cast the gpudata to int as it can either be a DeviceAllocation object
            # or an int (e.g. when using a view of another array)
            src_ptr = int(src.gpudata)
        if dest is not None:
            if use_cupy:
                dest_ptr = dest.__cuda_array_interface__['data'][0]
            else:
                dest_ptr = int(dest.gpudata)
        else:
            dest_ptr = src_ptr
        if self.inplace:
            if dest is not None:
                if src_ptr != dest_ptr:
                    raise RuntimeError("VkFFTApp.fft: dest!=src but this is an inplace transform")
            res = _vkfft_cuda.ifft(self.app, int(src_ptr), int(src_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst, backend="cuda")
            if self.norm == "ortho":
                src *= self._get_ifft_scale(norm=0)
            if self.r2c:
                if src.dtype == np.complex64:
                    return src.view(dtype=np.float32)
                elif src.dtype == np.complex128:
                    return src.view(dtype=np.float64)
            return src
        else:
            if dest is None:
                raise RuntimeError("VkFFTApp.ifft: dest is None but this is an out-of-place transform")
            if src_ptr == dest_ptr:
                raise RuntimeError("VkFFTApp.ifft: dest and src are identical but this is an out-of-place transform")
            if self.r2c:
                assert (src.size == dest.size // dest.shape[self.fast_axis] * (dest.shape[self.fast_axis] // 2 + 1))
                # Special case, src and dest buffer sizes are different,
                # VkFFT is configured to go back to the source buffer
                res = _vkfft_cuda.ifft(self.app, int(dest_ptr), int(src_ptr))
            else:
                res = _vkfft_cuda.ifft(self.app, int(src_ptr), int(dest_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst, backend="cuda")
            if self.norm == "ortho":
                dest *= self._get_ifft_scale(norm=0)
            return dest


def vkfft_version():
    """
    Get VkFFT version
    :return: version as X.Y.Z
    """
    int_ver = _vkfft_cuda.vkfft_version()
    return "%d.%d.%d" % (int_ver // 10000, (int_ver % 10000) // 100, int_ver % 100)


def vkfft_max_fft_dimensions():
    """
    Get the maximum number of dimensions VkFFT can handle. This is
    set at compile time. VkFFT default is 4, pyvkfft sets this to 8.
    Note that consecutive non-transformed are collapsed into a single
    axis, reducing the effective number of dimensions.

    :return: VKFFT_MAX_FFT_DIMENSIONS
    """
    return _vkfft_cuda.vkfft_max_fft_dimensions()


def cuda_runtime_version(raw=False):
    """
    Get CUDA runtime version

    :param raw: if True, return the version as X*1000+Y*10+Z
    :return: version as X.Y.Z
    """
    int_ver = _vkfft_cuda.cuda_runtime_version()
    if raw:
        return raw
    return "%d.%d.%d" % (int_ver // 1000, (int_ver % 1000) // 10, int_ver % 10)


def cuda_driver_version(raw=False):
    """
    Get CUDA driver version

    :param raw: if True, return the version as X*1000+Y*10+Z
    :return: version as X.Y.Z
    """
    int_ver = _vkfft_cuda.cuda_driver_version()
    if raw:
        return raw
    return "%d.%d.%d" % (int_ver // 1000, (int_ver % 1000) // 10, int_ver % 10)


def cuda_compile_version(raw=False):
    """
    Get CUDA version against which pyvkfft was compiled

    :param raw: if True, return the version as X*1000+Y*10+Z
    :return: version as X.Y.Z
    """
    if raw:
        return raw
    int_ver = _vkfft_cuda.cuda_compile_version()
    return "%d.%d.%d" % (int_ver // 1000, (int_ver % 1000) // 10, int_ver % 10)



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2021- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr
#
#
# Functions for accuracy tests.

import os
import platform
import multiprocessing
import timeit
import atexit

import psutil
import numpy as np
from numpy.fft import fftn, ifftn, rfftn, irfftn
from pyvkfft.base import primes, strides_nonzero

try:
    # We prefer scipy over numpy for fft, and we can also test dct
    from scipy.fft import dctn, idctn, fftn, ifftn, rfftn, irfftn, dstn, idstn

    has_dct_ref = True
    has_scipy = True
except ImportError:
    has_dct_ref = False
    has_scipy = False

# pyfftw speed is not good compared to scipy, when using every transform once
# try:
#     from pyfftw.interfaces import scipy_fft as pyfftw_fft
#     if not has_scipy:
#         from pyfftw.interfaces.scipy_fft import dctn, idctn, fftn, ifftn, rfftn, irfftn
#
#     has_pyfftw = True
#     has_dct_ref = True
# except ImportError:
#     has_pyfftw = False

try:
    import pyopencl as cl
    import pyopencl.array as cla
    from pyvkfft.opencl import VkFFTApp as clVkFFTApp

    has_opencl = True
except ImportError:
    has_opencl = False

try:
    from pyvkfft.cuda import VkFFTApp as cuVkFFTApp, has_pycuda, has_cupy

    if has_pycuda:
        import pycuda.driver as cu_drv
        import pycuda.gpuarray as cua

    if has_cupy:
        import cupy as cp
except ImportError:
    has_cupy = False
    has_pycuda = False

# Dictionary of cuda/opencl (device, context). Will be initialised on-demand.
# This is needed for multiprocessing.
# The pyopencl entry is a tuple with (device, context, queue, has_cl_fp64)
gpu_ctx_dic = {}


def init_ctx(backend, gpu_name=None, opencl_platform=None, verbose=False):
    if backend in gpu_ctx_dic:
        return
    if backend == "pycuda":
        if not has_pycuda:
            raise RuntimeError("init_ctx: backend=%s is not available" % backend)
        cu_drv.init()
        d = None
        if gpu_name is not None:
            for i in range(cu_drv.Device.count()):
                if gpu_name.lower() in cu_drv.Device(i).name().lower():
                    d = cu_drv.Device(i)
                    break
        else:
            d = cu_drv.Device(0)
        if d is None:
            if gpu_name is not None:
                raise RuntimeError("Selected backend is pycuda, but no device found (name=%s)" % gpu_name)
            else:
                raise RuntimeError("Selected backend is pycuda, but no device found")
        gpu_ctx_dic["pycuda"] = (d, d.retain_primary_context())
        gpu_ctx_dic["pycuda"][1].push()
        if verbose:
            print("Selected device for pycuda: %s" % d.name())
    elif backend == "pyopencl":
        if not has_opencl:
            raise RuntimeError("init_ctx: backend=%s is not available" % backend)
        d = None
        for p in cl.get_platforms():
            if d is not None:
                break
            if opencl_platform is not None:
                if opencl_platform.lower() not in p.name.lower():
                    continue
            for d0 in p.get_devices():
                if d0.type & cl.device_type.GPU:
                    if gpu_name is not None:
                        if gpu_name.lower() in d0.name.lower():
                            d = d0
                    else:
                        d = d0
                if d is not None:
                    break
        if d is None:
            raise RuntimeError("Selected backend is pyopencl, but no device found (name=%s, platform=%s)" %
                               (gpu_name, opencl_platform))
        cl_ctx = cl.Context([d])
        cq = cl.CommandQueue(cl_ctx)
        gpu_ctx_dic["pyopencl"] = d, cl_ctx, cq, 'cl_khr_fp64' in cq.device.extensions
        if verbose:
            print("Selected device for pyopencl: %s [%s]" % (d.name, p.name))
    elif backend == "cupy":
        if not has_cupy:
            raise RuntimeError("init_ctx: backend=%s is not available" % backend)
        if gpu_name is not None:
            for i in range(cp.cuda.runtime.getDeviceCount()):
                if gpu_name.lower() in cp.cuda.runtime.getDeviceProperties(i)['name'].decode().lower():
                    d = cp.cuda.Device(i).use()
                    break
        else:
            d = cp.cuda.Device(0).use()
        if d is None:
            if gpu_name is not None:
                raise RuntimeError("Selected backend is pycuda, but no device found (name=%s)" % gpu_name)
            else:
                raise RuntimeError("Selected backend is pycuda, but no device found")
        gpu_ctx_dic["cupy"] = d

        # TODO: The following somehow helps initialising cupy, not sure why it's useful.
        #  (some context auto-init...). Otherwise a cuLaunchKernel error occurs with
        #  the first transform.
        cupy_a = cp.array(np.zeros((128, 128), dtype=np.float32))
        cupy_a.sum()
    else:
        raise RuntimeError("init_ctx: unknown backend ", backend)


def cleanup_cu_ctx():
    # Is that really clean ?
    if has_pycuda:
        if cu_drv.Context is not None:
            while cu_drv.Context.get_current() is not None:
                cu_drv.Context.pop()


atexit.register(cleanup_cu_ctx)


def l2(a, b):
    """L2 norm"""
    return np.sqrt((abs(a - b) ** 2).sum() / (abs(a) ** 2).sum())


def li(a, b):
    """Linf norm"""
    return abs(a - b).max() / abs(a).max()


def test_accuracy(backend, shape, ndim, axes, dtype, inplace, norm, use_lut,
                  r2c=False, dct=False, dst=False,
                  gpu_name=None, opencl_platform=None, stream=None, queue=None, return_array=False,
                  init_array=None, verbose=False, colour_output=False, ref_long_double=True, order='C'):
    """
    Measure the FT accuracy by comparing to the result from scipy (if available), or numpy.

    :param backend: either 'pyopencl', 'pycuda' or 'cupy'
    :param shape: the shape of the array to test. If this is an inplace r2c, the
        fast-axis length must be even, and two extra values will be appended along x,
        so the actual transform shape is the one supplied
    :param ndim: the number of FFT dimensions. Can be None if axes is given
    :param axes: the transform axes. Supersedes ndim
    :param dtype: either np.complex64 or np.complex128, or np.float32/np.float64 for r2c & dct
    :param inplace: if True, make an inplace transform. Note that for inplace r2c transforms,
        the size for the last (x, fastest) axis must be even.
    :param norm: either 0, 1 or "ortho"
    :param use_lut: if True,1, False or 0, will trigger useLUT=1 or 0 for VkFFT.
        If None, the default VkFFT behaviour is used.
    :param r2c: if True, test an r2c transform. If inplace, the last dimension
        (x, fastest axis) must be even
    :param dct: either 1, 2, 3 or 4 to test different dct. Only norm=1 is can be
        tested (native scipy normalisation).
    :param dst: either 1, 2, 3 or 4 to test different dst. Only norm=1 is can be
        tested (native scipy normalisation).
    :param gpu_name: the name of the gpu to use. If None, the first available
        for the backend will be used.
    :param opencl_platform: the name of the OpenCL platform to use. If None, the first available
        will be used.
    :param stream: the cuda stream to use, or None
    :param queue: the opencl queue to use (mandatory for the 'pyopencl' backend)
    :param return_array: if True, will return the generated random array so it can be
        re-used for different parameters
    :param init_array: the initial  (numpy) random array to use (should be filled
        with uniform random numbers between +/-0.5 for both real and imaginary
        fields), to save time. The correct type will be applied.
        If None, a random array is generated.
    :param verbose: if True, print a 1-line info for both fft and ifft results
    :param colour_output: if True, use some colour to tag the quality of the accuracy
    :param ref_long_double: if True and scipy is available, long double precision
        will be used for the reference transform. Otherwise, this is ignored.
    :param order: either 'C' (default C-contiguous) or 'F' to test a different
        stride. Note that for the latter, a 3D transform on a 4D array will not
        be supported as the last transform axis would be on the 4th dimension
        (once ordered by stride).
    :return: a dictionary with (l2_fft, li_fft, l2_ifft, li_ifft, tol, dt_array,
        dt_app, dt_fft, dt_ifft, src_unchanged_fft, src_unchanged_ifft, tol_test, str),
        with the L2 and Linf normalised norms comparing pyvkfft's result with either
        numpy, scipy, the reference
        tolerance, and the times spent in preparing the initial random array, creating
        the VkFFT app, and performing the forward and backward transforms (including
        the GPU and reference transforms, plus the L2 and Linf computations - don't use
        this for benchmarking), 'src_fft_unchanged' and 'srf_ifft_unchanged' are True
        if for an out-of-place transform, the source array is actually unmodified
        (which is not true for r2c ifft with ndim>=2).
        The last fields are 'tol_test' which is True if both li_fft and li_ifft are
        smaller than tol, and str the string summarising the results (printed if
        verbose is True). If return_array is True, the initial random array used is
        returned as 'd0'.
        All input parameters are also returned as key/values, except stream, queue,
        return_array, ini_array and verbose.
    """
    ndims = len(shape)
    if backend == "cupy" and has_cupy:
        mempool = cp.get_default_memory_pool()
        if mempool is not None:  # Is that test necessary ?
            # Clean memory pool, we are changing array sizes constantly, and using
            # N parallel process so memory management  must be done manually
            mempool.free_all_blocks()
    t0 = timeit.default_timer()
    init_ctx(backend, gpu_name=gpu_name, opencl_platform=opencl_platform, verbose=False)
    if backend == "pyopencl" and queue is None:
        queue = gpu_ctx_dic["pyopencl"][2]
    shape0 = shape
    dtype0 = dtype
    if dtype in (np.complex64, np.float32):
        dtype = np.complex64
        dtypef = np.float32
    else:
        dtype = np.complex128
        dtypef = np.float64

    if dct or dst:
        if norm != 1:
            raise RuntimeError("test_accuracy: only norm=1 can be used with dct or dst")

    # Need the fast axis for R2C (last for 'C' order, first for 'F')
    # Careful with special cases of axes with size 1, where stride does not increase or is 0...
    fast_axis = 0 if order == 'F' else ndims - 1

    if r2c:
        r2c_odd = shape[-1] % 2 == 1 if order == 'C' else shape[0] % 2 == 1
        r2c_inplace_pad = 1 if r2c_odd else 2
        if inplace:
            # Add two (or one for an odd-sized fast axis) extra columns in the source array
            # so the transform has the desired shape
            shape = list(shape)
            if order == 'C':
                shape[-1] += r2c_inplace_pad
            else:
                shape[0] += r2c_inplace_pad
        else:
            shapec = list(shape)
            shapec[fast_axis] = shapec[fast_axis] // 2 + 1
            shapec = tuple(shapec)
    else:
        r2c_odd = None
        r2c_inplace_pad = None
        shapec = tuple(shape)
    shape = tuple(shape)

    if init_array is not None:
        if r2c:
            if inplace:
                d0 = np.empty(shape, dtype=dtypef)
                if order == 'C':
                    d0[..., :-r2c_inplace_pad] = init_array
                else:
                    d0[:-r2c_inplace_pad] = init_array
            else:
                d0 = init_array.astype(dtypef)
        elif dct or dst:
            d0 = init_array.astype(dtypef)
        else:
            d0 = init_array.astype(dtype)
    else:
        rng = np.random.default_rng(seed=None)
        if r2c or dct or dst:
            d0 = rng.uniform(-0.5, 0.5, size=shape).astype(dtypef)
        else:
            d0 = (rng.uniform(-0.5, 0.5, size=shape) + 1j * rng.uniform(-0.5, 0.5, size=shape)).astype(dtype)

    if order != 'C':
        d0 = np.asarray(d0, order=order)

    t1 = timeit.default_timer()

    if 'opencl' in backend:
        app = clVkFFTApp(d0.shape, d0.dtype, queue, ndim=ndim, norm=norm,
                         axes=axes, useLUT=use_lut, inplace=inplace,
                         r2c=r2c, dct=dct, dst=dst, strides=d0.strides,
                         r2c_odd=r2c_odd)
        t2 = timeit.default_timer()
        d_gpu = cla.to_device(queue, d0)
        empty_like = cla.empty_like
    else:
        if backend == "pycuda":
            to_gpu = cua.to_gpu
            empty_like = cua.empty_like
        else:
            to_gpu = cp.array
            empty_like = cp.empty_like

        app = cuVkFFTApp(d0.shape, d0.dtype, ndim=ndim, norm=norm, axes=axes,
                         useLUT=use_lut, inplace=inplace, r2c=r2c,
                         dct=dct, dst=dst, strides=d0.strides,
                         r2c_odd=r2c_odd, stream=stream)
        t2 = timeit.default_timer()
        d_gpu = to_gpu(d0)

    if axes is None:
        axes_numpy = list(range(ndims))[-ndim:] if order == 'C' else list(range(ndim))
    else:
        # Make sure axes indices are >0
        axes_numpy = [ax if ax >= 0 else ndims + ax for ax in axes]

    if r2c:
        if fast_axis not in axes_numpy:
            raise RuntimeError(f"The fast axis must be transformed for R2C: "
                               f"fast axis={fast_axis}, axes_numpy={axes_numpy}")

        # For R2C, we need the same fast axis as on the GPU, or the
        # half-hermitian result won't look the same
        if fast_axis != axes_numpy[-1]:
            axes_numpy.remove(fast_axis)
            axes_numpy.append(fast_axis)
        # if order != 'C':
        #     print("R2C", ndims, ndim, shape, axes, axes_numpy, fast_axis, inplace)

    # base FFT scale for numpy (not used for DCT/DST)
    s = np.sqrt(np.prod([d0.shape[i] for i in axes_numpy]))
    if r2c and inplace:
        s = np.sqrt(s ** 2 / d0.shape[fast_axis] * (d0.shape[fast_axis] - r2c_inplace_pad))

    # Tolerance estimated from accuracy notebook
    if dtype in (np.complex64, np.float32):
        tol = 2e-6 + 5e-7 * np.log10(s ** 2)
    else:
        tol = 5e-15 + 5e-16 * np.log10(s ** 2)

    n = max(shape)
    bluestein = max(primes(n)) > 13
    if bluestein:
        tol *= 2
    ############################################################
    # FFT
    ############################################################
    if inplace:
        d1_gpu = d_gpu
    else:
        if r2c:
            if backend == "pyopencl":
                d1_gpu = cla.empty(queue, shapec, dtype=dtype, order=order)
            elif backend == "pycuda":
                d1_gpu = cua.empty(shapec, dtype=dtype, order=order)
            elif backend == "cupy":
                d1_gpu = cp.empty(shapec, dtype=dtype, order=order)
        else:
            # Note that pycuda's gpuarray.copy (as of 2022.2.2) does not copy strides
            d1_gpu = empty_like(d_gpu)

    if has_scipy and ref_long_double:
        # Use long double precision
        if r2c or dct or dst:
            d0n = d0.astype(np.longdouble)
        else:
            d0n = d0.astype(np.clongdouble)
    else:
        d0n = d0

    d1_gpu = app.fft(d_gpu, d1_gpu)
    if not (dct or dst):
        d1_gpu *= app.get_fft_scale()

    if r2c:
        if inplace:
            # Need to cut the fastest axis by 1 or 2
            d = rfftn(np.take(d0n, range(d0n.shape[fast_axis] - r2c_inplace_pad),
                              axis=fast_axis), axes=axes_numpy) / s
        else:
            d = rfftn(d0n, axes=axes_numpy) / s
    elif dct:
        d = dctn(d0n, axes=axes_numpy, type=dct)
    elif dst:
        d = dstn(d0n, axes=axes_numpy, type=dst)
    else:
        d = fftn(d0n, axes=axes_numpy) / s

    if inplace and r2c:
        assert d1_gpu.dtype == dtype, "The array type is incorrect after an inplace FFT"

    if r2c and d.shape != d1_gpu.shape:
        raise RuntimeError(f"Shape mismatch after the R2C FFT: shape={shape} shapec={shapec} "
                           f"order={order} axes={axes} axes_numpy={axes_numpy} ndim={ndim}")
    n2, ni = l2(d, d1_gpu.get()), li(d, d1_gpu.get())

    src_unchanged_fft = np.all(np.equal(d_gpu.get(), d0))

    # Output string
    if r2c:
        t = "R2C"
    elif dct:
        t = "DCT%d" % dct
    elif dst:
        t = "DST%d" % dst
    else:
        t = "C2C"
    # if r2c and inplace:
    #     tmp = list(d0.shape)
    #     if order == 'C':
    #         tmp[-1] -= r2c_inplace_pad
    #         shstr = str(tuple(tmp)).replace(" ", "")
    #         if ",)" in shstr:
    #             shstr = shstr.replace(",)", f"+{r2c_inplace_pad})")
    #         else:
    #             shstr = shstr.replace(")", f"+{r2c_inplace_pad})")
    #     else:
    #         tmp[0] -= r2c_inplace_pad
    #         if len(tmp) == 1:
    #             shstr = f"({tmp[0]}+{r2c_inplace_pad}),"
    #         else:
    #             shstr = f"({tmp[0]}+{r2c_inplace_pad},"
    #             shstr += str(tuple(tmp[1:])).replace(" ", "").replace(",)", ")")[1:]
    # else:
    #     shstr = str(d0.shape).replace(" ", "")
    shstr = app.get_shape_str()
    shax = str(axes).replace(" ", "").replace("[", "").replace("]", "")
    if colour_output:
        red = max(0, min(int((ni / tol - 0.2) * 255), 255))
        stol = "\x1b[48;2;%d;0;0m%5.1e < %5.1e (%5.3f)\x1b[0m" % (red, ni, tol, ni / tol)
    else:
        stol = "%5.1e < %5.1e (%5.3f)" % (ni, tol, ni / tol)
    nupstr = ''.join(str(nup) for nup in app.get_nb_upload())
    verb_out = "%8s %4s %15s axes=%12s ndim=%4s %5s %5s %10s lut=%4s inplace=%d " \
               " norm=%4s %s %5s: n2=%5.1e ninf=%s %d" % \
               (backend, t, shstr, shax, str(ndim), app.get_algo_str(), nupstr, str(d0.dtype),
                str(use_lut), int(inplace), str(norm), order, "FFT", n2, stol, src_unchanged_fft)

    t3 = timeit.default_timer()

    # Clean memory
    del d_gpu, d1_gpu
    if backend == "cupy" and has_cupy:
        mempool = cp.get_default_memory_pool()
        if mempool is not None:  # Is that test necessary ?
            # Clean memory pool, we are changing array sizes constantly, and using
            # N parallel process so memory management  must be done manually
            mempool.free_all_blocks()

    ############################################################
    # IFFT - from original array to avoid error propagation
    ############################################################
    if r2c:
        # Exception: we need a proper half-Hermitian array
        d0 = d.astype(dtype)
        if has_scipy and ref_long_double:
            d0n = d0.astype(np.clongdouble)
        else:
            d0n = d0
        if (np.argmin(strides_nonzero(d0.strides)[::-1]) != 0) and order == 'C':
            # np.fft.rfftn can change the fast axis
            d0 = np.asarray(d0, order='C')
    if order != 'C':
        d0 = np.asarray(d0, order=order)
    if 'opencl' in backend:
        d_gpu = cla.to_device(queue, d0)
    else:
        d_gpu = to_gpu(d0)

    if inplace:
        d1_gpu = d_gpu
    else:
        if r2c:
            if backend == "pyopencl":
                d1_gpu = cla.empty(queue, shape, dtype=dtypef, order=order)
            elif backend == "pycuda":
                d1_gpu = cua.empty(shape, dtype=dtypef, order=order)
            elif backend == "cupy":
                d1_gpu = cp.empty(shape, dtype=dtypef, order=order)
        else:
            d1_gpu = empty_like(d_gpu)

    d1_gpu = app.ifft(d_gpu, d1_gpu)
    if not (dct or dst):
        d1_gpu *= app.get_ifft_scale()

    if r2c:
        # We need to specify the destination shape, in the case
        # we want an odd-sized fast axis
        d = irfftn(d0n, s=[shape0[i] for i in axes_numpy], axes=axes_numpy) * s
    elif dct:
        d = idctn(d0n, axes=axes_numpy, type=dct)
    elif dst:
        d = idstn(d0n, axes=axes_numpy, type=dst)
    else:
        d = ifftn(d0n, axes=axes_numpy) * s

    if inplace:
        if dct or dst or r2c:
            assert d1_gpu.dtype == dtypef, "The array type is incorrect after an inplace iFFT"
        else:
            assert d1_gpu.dtype == dtype, "The array type is incorrect after an inplace iFFT"

    if r2c and inplace:
        tmp = np.take(d1_gpu.get(), range(d1_gpu.shape[fast_axis] - r2c_inplace_pad), axis=fast_axis)
        n2i, nii = l2(d, tmp), li(d, tmp)
    else:
        n2i, nii = l2(d, d1_gpu.get()), li(d, d1_gpu.get())

    src_unchanged_ifft = np.all(np.equal(d_gpu.get(), d0))

    # Max N for radix 1D C2R transforms to not overwrite source
    if platform.system() == 'Darwin':
        nmaxr2c1d = 2048 * (1 + int(dtype in (np.float32, np.complex64)))
    else:
        nmaxr2c1d = 3072 * (1 + int(dtype in (np.float32, np.complex64)))
    ndimtmp = ndim if ndim is not None else len(axes)
    if max(ni, nii) <= tol and (inplace or src_unchanged_fft) and \
            (inplace or src_unchanged_ifft or (r2c and ndimtmp > 1 or n >= nmaxr2c1d or bluestein)):
        success = 'OK'
    else:
        success = 'FAIL'

    if colour_output:
        red = max(0, min(int((nii / tol - 0.2) * 255), 255))
        stol = "\x1b[48;2;%d;0;0m%5.1e < %5.1e (%5.3f)\x1b[0m" % (red, nii, tol, nii / tol)
    else:
        stol = "%5.1e < %5.1e (%5.3f)" % (nii, tol, nii / tol)
    verb_out += "%5s: n2=%5.1e ninf=%s %d" % ("iFFT", n2i, stol, src_unchanged_ifft)

    # Also print the size of the allocated buffer and success
    verb_out += f" buf={app.get_tmp_buffer_str()} {success:4s}"

    if verbose:
        print(verb_out)

    t4 = timeit.default_timer()

    if backend == "pyopencl":
        gpu_name = gpu_ctx_dic["pyopencl"][0].name
    elif backend == "pycuda":
        gpu_name = gpu_ctx_dic["pycuda"][0].name()
    else:
        gpu_name = ""

    res = {"n2": n2, "ni": ni, "n2i": n2i, "nii": nii, "tol": tol, "dt_array": t1 - t0, "dt_app": t2 - t1,
           "dt_fft": t3 - t2, "dt_ifft": t4 - t3, "src_unchanged_fft": src_unchanged_fft,
           "src_unchanged_ifft": src_unchanged_ifft, "tol_test": max(ni, nii) < tol, "str": verb_out,
           "backend": backend, "shape": shape0, "ndim": ndim, "axes": axes, "dtype": dtype0, "inplace": inplace,
           "norm": norm, "use_lut": use_lut, "r2c": r2c, "dct": dct, "dst": dst,
           "gpu_name": gpu_name, "order": order}

    if return_array:
        res["d0"] = d0
    return res


def test_accuracy_kwargs(kwargs):
    # This function must be defined here, so it can be used with a multiprocessing pool
    # in test_fft, otherwise this will fail, see:
    # https://stackoverflow.com/questions/41385708/multiprocessing-example-giving-attributeerror
    if kwargs['backend'] == 'pyopencl' and has_opencl:
        try:
            t = test_accuracy(**kwargs)
        except cl.RuntimeError as ex:
            # The cl.RuntimeError can't be pickled, so is not correctly reported
            # when using multiprocessing. So we raise another and the traceback
            # from the previous one is still printed.
            raise RuntimeError("An OpenCL RuntimeError was encountered")
        return t
    return test_accuracy(**kwargs)


def exhaustive_test(backend, vn, ndim, dtype, inplace, norm, use_lut, r2c=False,
                    dct=False, dst=False, nproc=None,
                    verbose=True, return_res=False):
    """
    Run tests on a large range of sizes using multiprocessing. Manual function.

    :param backend: either 'pyopencl', 'pycuda' or 'cupy'
    :param vn: the list/iterable of sizes n.
    :param ndim: the number of dimensions. The array shape will be [n]*ndim
    :param dtype: either np.complex64 or np.complex128, or np.float32/np.float64 for r2c/dct/dst
    :param inplace: True or False
    :param norm: either 0, 1 or "ortho"
    :param use_lut: if True,1, False or 0, will trigger useLUT=1 or 0 for VkFFT.
        If None, the default VkFFT behaviour is used. Always True by default
        for double precision, so no need to force it.
    :param r2c: if True, test an r2c transform. If inplace, the last dimension
        (x, fastest axis) must be even
    :param dct: either 1, 2, 3 or 4 to test different dct. Only norm=1 is can be
        tested (native scipy normalisation).
    :param dst: either 1, 2, 3 or 4 to test different dst. Only norm=1 is can be
        tested (native scipy normalisation).
    :param nproc: the maximum number of parallel process to use. If None, the
        number of detected cores will be used (this may use too much memory !)
    :param verbose: if True, prints 1 line per test
    :param return_res: if True, return the list of result dictionaries.
    :return: True if all tests passed, False otherwise. If return_res is True, return
        the list of result dictionaries instead.
    """
    try:
        # Get the real number of processor cores available
        # os.sched_getaffinity is only available on some *nix platforms
        nproc1 = len(os.sched_getaffinity(0)) * psutil.cpu_count(logical=False) // psutil.cpu_count(logical=True)
    except AttributeError:
        nproc1 = os.cpu_count()
    if nproc is None:
        nproc = nproc1
    else:
        nproc = min(nproc, nproc1)
    # Generate the list of configurations as kwargs for test_accuracy()
    vkwargs = []
    for n in vn:
        kwargs = {"backend": backend, "shape": [n] * ndim, "ndim": ndim, "axes": None, "dtype": dtype,
                  "inplace": inplace, "norm": norm, "use_lut": use_lut,
                  "r2c": r2c, "dct": dct, "dst": dst, "stream": None,
                  "verbose": False}
        vkwargs.append(kwargs)
    vok = []
    vres = []
    # Need to use spawn to handle the GPU context
    with multiprocessing.get_context('spawn').Pool(nproc) as pool:
        for res in pool.imap(test_accuracy_kwargs, vkwargs):
            # TODO: this should better be logged
            if verbose:
                print(res['str'])
            ni, n2 = res["ni"], res["n2"]
            nii, n2i = res["nii"], res["n2i"]
            tol = res["tol"]
            ok = max(ni, nii) < tol
            if not inplace:
                ok = ok and res["src_unchanged_fft"]
                if not r2c:
                    ok = ok and res["src_unchanged_ifft"]
            vok.append(ok)
            if return_res:
                vres.append(res)
    if return_res:
        return vres
    return np.all(vok)



# The setup used here is derived with bits from:
# - https://github.com/rmcgibbo/npcuda-example

import os
import sys
import platform
import subprocess
from os.path import join as pjoin
import warnings
from setuptools import setup, find_packages
from setuptools.command.sdist import sdist
from distutils.extension import Extension
from distutils import unixccompiler
from setuptools.command.build_ext import build_ext as build_ext_orig
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.install_lib import install_lib as su_install_lib
from setuptools.command.sdist import sdist as su_sdist
from pyvkfft.version import __version__, git_version, vkfft_git_version

# Maximum number of dimensions VkFFT can handle. VkFFT sets this to 4,
# pyvkfft uses a default of 8. Set an environment variable
# VKFFT_MAX_FFT_DIMENSIONS to increase this
if 'VKFFT_MAX_FFT_DIMENSIONS' in os.environ:
    VKFFT_MAX_FFT_DIMENSIONS = os.environ['VKFFT_MAX_FFT_DIMENSIONS']
else:
    VKFFT_MAX_FFT_DIMENSIONS = 8


def find_in_path(name, path):
    """Find a file in a search path"""

    # Adapted fom http://code.activestate.com/recipes/52224
    for dir in path.split(os.pathsep):
        binpath = pjoin(dir, name)
        if os.path.exists(binpath):
            return os.path.abspath(binpath)
    return None


def locate_cuda():
    """Locate the CUDA environment on the system

    Returns a dict with keys 'home', 'nvcc', 'include', and 'lib64'
    and values giving the absolute path to each directory.

    Starts by looking for the CUDAHOME or CUDA_PATH env variable.
    If not found, find 'nvcc' in the PATH.
    """
    # search for nvcc
    app_name = 'nvcc.exe' if platform.system() == "Windows" else 'nvcc'
    nvcc, home = None, None
    for p in ['CUDA_PATH', 'CUDA_HOME', 'CUDAHOME']:
        if p in os.environ:
            if os.path.exists(pjoin(os.environ[p], 'bin', app_name)):
                nvcc = pjoin(os.environ[p], 'bin', app_name)
                home = os.environ[p]
                break
    if nvcc is None:
        # Search the PATH for NVCC
        nvcc = find_in_path(app_name, os.environ['PATH'])
        if nvcc is not None:
            home = os.path.dirname(os.path.dirname(nvcc))
        else:
            raise EnvironmentError('The nvcc binary could not be '
                                   'located in your $PATH. Either add it to your path, '
                                   'or set $CUDA_PATH')

    if platform.system() == "Windows":
        libdir = pjoin(home, 'lib', 'x64')
        extra_compile_args = ['-O3', '--ptxas-options=-v', '-Xcompiler', '-MD',
                              f'-DVKFFT_MAX_FFT_DIMENSIONS={VKFFT_MAX_FFT_DIMENSIONS}',
                              "--use-local-env", "-ccbin",
                              os.path.dirname(find_in_path('cl.exe', os.environ['PATH']))]
        tmp = pjoin(os.path.dirname(find_in_path('cl.exe', os.environ['PATH'])).split('bin')[0],
                    'lib', 'x64')
        extra_link_args = ['-L%s' % libdir, "--use-local-env", "-ccbin",
                           os.path.dirname(find_in_path('cl.exe', os.environ['PATH'])),
                           '-L%s' % tmp]
    else:
        if os.path.exists(pjoin(home, 'lib64')):
            libdir = pjoin(home, 'lib64')
        else:
            libdir = pjoin(home, 'lib')
        extra_compile_args = ['-O3', '--ptxas-options=-v', '-std=c++14',
                              '--compiler-options=-fPIC',
                              f'-DVKFFT_MAX_FFT_DIMENSIONS={VKFFT_MAX_FFT_DIMENSIONS}']
        extra_link_args = ['--shared', '-L%s' % libdir]
    cudaconfig = {'home': home, 'nvcc': nvcc,
                  'include_dirs': [pjoin(home, 'include'), 'src/VkFFT/vkFFT'],
                  'extra_compile_args': extra_compile_args,
                  'extra_link_args': extra_link_args}
    if 'CONDA_PREFIX' in os.environ:
        cudaconfig['include_dirs'].append(f"{os.environ['CONDA_PREFIX']}/include")
        cudaconfig['include_dirs'].append(f"{os.environ['CONDA_PREFIX']}/Library/include")
    return cudaconfig


def locate_opencl():
    """
    Get the opencl configuration
    :return:
    """
    include_dirs = ['src/VkFFT/vkFFT']
    library_dirs = []
    extra_compile_args = ['-std=c++14', '-Wno-format-security',
                          f'-DVKFFT_MAX_FFT_DIMENSIONS={VKFFT_MAX_FFT_DIMENSIONS}']
    extra_link_args = None
    if platform.system() == 'Darwin':
        libraries = None
        extra_link_args = ['-Wl,-framework,OpenCL']
    elif platform.system() == "Windows":
        # Add include & lib dirs if possible from usual nvidia and AMD paths
        for path in ["CUDA_HOME", "CUDAHOME", "CUDA_PATH"]:
            if path in os.environ:
                include_dirs.append(pjoin(os.environ[path], 'include'))
                library_dirs.append(pjoin(os.environ[path], 'lib', 'x64'))
        library_dirs.append(pjoin(os.path.dirname(find_in_path('cl.exe', os.environ['PATH'])).split('bin')[0],
                                  'lib', 'x64'))
        libraries = ['OpenCL']
        extra_compile_args = [f'/DVKFFT_MAX_FFT_DIMENSIONS={VKFFT_MAX_FFT_DIMENSIONS}']
    else:
        # Linux
        libraries = ['OpenCL']
        extra_link_args = ['--shared']

    if 'CONDA_PREFIX' in os.environ:
        include_dirs.append(f"{os.environ['CONDA_PREFIX']}/include")
        include_dirs.append(f"{os.environ['CONDA_PREFIX']}/Library/include")
        library_dirs.append(f"{os.environ['CONDA_PREFIX']}/Library/Lib")

    opencl_config = {'libraries': libraries, 'extra_link_args': extra_link_args,
                     'include_dirs': include_dirs, 'library_dirs': library_dirs,
                     'extra_compile_args': extra_compile_args}
    return opencl_config


class build_ext_custom(build_ext_orig):
    """Custom `build_ext` command which will correctly compile and link
    the OpenCL and CUDA modules. The hooks are based on the name of the extension"""

    def build_extension(self, ext):
        if "cuda" in ext.name:
            # Use nvcc for compilation. This assumes all sources are .cu files
            # for this extension.
            default_compiler = self.compiler
            # Create unix compiler patched for cu
            self.compiler = unixccompiler.UnixCCompiler()
            self.compiler.src_extensions.append('.cu')
            tmp = CUDA['nvcc']  # .replace('\\\\','toto')
            self.compiler.set_executable('compiler_so', [tmp])
            self.compiler.set_executable('linker_so', [tmp])
            if platform.system() == "Windows":
                CUDA['extra_link_args'] += ['--shared', '-Xcompiler', '-MD']
                # pythonXX.lib must be in the linker paths
                # Is using sys.prefix\libs always correct ?
                CUDA['extra_link_args'].append('-L%s' % pjoin(sys.prefix, 'libs'))

            super().build_extension(ext)
            # Restore default linker and compiler
            self.compiler = default_compiler
        else:
            super().build_extension(ext)

    def get_export_symbols(self, ext):
        """ Hook to make sure we get the correct symbols for windows"""
        if ("opencl" in ext.name or "cuda" in ext.name) and platform.system() == "Windows":
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        """ Hook to make sure we keep the correct name (*.so) for windows"""
        if ("opencl" in ext_name or "cuda" in ext_name) and platform.system() == "Windows":
            return ext_name + '.so'
        return super().get_ext_filename(ext_name)


ext_modules = []
install_requires = ['numpy', 'psutil']
exclude_packages = ['examples']
CUDA = None
OPENCL = None

if "VKFFT_BACKEND" in os.environ:
    v = os.environ["VKFFT_BACKEND"]
    # Environment variable to manually select vkfft backends. useful e.g. if nvidia tools
    # are installed but not functional
    # e.g. use:
    #   VKFFT_BACKEND=opencl pip install pyvkfft
    #   VKFFT_BACKEND=cuda pip install .
    if 'opencl' not in v.lower():
        exclude_packages.append('opencl')
    if 'cuda' not in v.lower():
        exclude_packages.append('cuda')

if 'cuda' not in exclude_packages:
    try:
        CUDA = locate_cuda()
        vkfft_cuda_ext = Extension('pyvkfft._vkfft_cuda',
                                   sources=['src/vkfft_cuda.cu'],
                                   libraries=['nvrtc', 'cuda'],
                                   extra_compile_args=CUDA['extra_compile_args'],
                                   include_dirs=CUDA['include_dirs'] + ['src'],
                                   extra_link_args=CUDA['extra_link_args'],
                                   depends=['vkFFT.h']
                                   )
        ext_modules.append(vkfft_cuda_ext)
        # install_requires.append("pycuda")
        try:
            import pycuda

            has_pycuda = True
        except ImportError:
            has_pycuda = False
        try:
            import cupy
        except ImportError:
            if has_pycuda is False:
                print("Reminder: you need to install either PyCUDA or CuPy to use pyvkfft.cuda")
    except:
        exclude_packages.append('cuda')
        warnings.warn("CUDA not available ($CUDAHOME/$CUDA_PATH variables missing "
                      "and nvcc not in path. "
                      "Skipping pyvkfft.cuda module installation.", UserWarning)

if 'opencl' not in exclude_packages:
    OPENCL = locate_opencl()
    install_requires.append('pyopencl')

    # OpenCL extension
    vkfft_opencl_ext = Extension('pyvkfft._vkfft_opencl',
                                 sources=['src/vkfft_opencl.cpp'],
                                 extra_compile_args=OPENCL['extra_compile_args'],
                                 include_dirs=OPENCL['include_dirs'] + ['src'],
                                 libraries=OPENCL['libraries'],
                                 library_dirs=OPENCL['library_dirs'],
                                 extra_link_args=OPENCL['extra_link_args'],
                                 depends=['vkFFT.h']
                                 )

    ext_modules.append(vkfft_opencl_ext)

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()


class bdist_egg_disabled(bdist_egg):
    """ Disabled bdist_egg, to prevent use of 'python setup.py install' """

    def run(self):
        sys.exit("Aborting building of eggs. Please use `pip install .` to install from source.")


# Console scripts, available e.g. as 'pyvkfft-test'
scripts = ['pyvkfft/scripts/pyvkfft_test.py', 'pyvkfft/scripts/pyvkfft_test_suite.py',
           'pyvkfft/scripts/pyvkfft_benchmark.py', 'pyvkfft/scripts/pyvkfft_info.py']

console_scripts = []
for s in scripts:
    s1 = os.path.splitext(os.path.split(s)[1])[0]
    s0 = os.path.splitext(s)[0]
    console_scripts.append("%s = %s:main" % (s1.replace('_', '-'), s0.replace('/', '.')))


class pyvkfft_sdist(su_sdist):
    """Hook to include git version of pyvkfft and VkFFT"""

    def make_release_tree(self, base_dir, files):
        super(pyvkfft_sdist, self).make_release_tree(base_dir, files)
        try:
            # Replace git_version_placeholder by real git version
            version_file = os.path.join(base_dir, "pyvkfft/version.py")
            vers = open(version_file).read()
            os.remove(version_file)
            with open(version_file, "w") as fh:
                vers = vers.replace("vkfft_git_version_placeholder", vkfft_git_version())
                vers = vers.replace("git_version_placeholder", git_version())
                fh.write(vers)
        except:
            print("sdist: replacing git_version failed")


class pyvkfft_install_lib(su_install_lib):
    """Hook to include git version of pyvkfft and VkFFT"""

    def run(self):
        super(pyvkfft_install_lib, self).run()
        try:
            # print(self.install_dir, self.build_dir)
            # Replace git_version_placeholder by real git version
            version_file = os.path.join(self.install_dir, "pyvkfft/version.py")
            vers = open(version_file).read()
            os.remove(version_file)
            with open(version_file, "w") as fh:
                vers = vers.replace("vkfft_git_version_placeholder", vkfft_git_version())
                vers = vers.replace("git_version_placeholder", git_version())
                fh.write(vers)
        except:
            print("install_lib: replacing git_version failed")


setup(name="pyvkfft",
      version=__version__,
      description="Python wrapper for the CUDA and OpenCL backends of VkFFT,"
                  "providing GPU FFT for PyCUDA, PyOpenCL and CuPy",
      long_description=long_description,
      ext_modules=ext_modules,
      packages=find_packages(exclude=exclude_packages),
      include_package_data=True,
      author="Vincent Favre-Nicolin",
      author_email="favre@esrf.fr",
      url="https://github.com/vincefn/pyvkfft",
      project_urls={
          "Source": "https://github.com/vincefn/pyvkfft",
          "Documentation": "https://pyvkfft.readthedocs.io/",
          "Bug Tracker": "https://github.com/vincefn/pyvkfft/issues",
          "VkFFT project": "https://github.com/DTolm/VkFFT",
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Environment :: GPU",
      ],
      license='MIT License',
      cmdclass={'build_ext': build_ext_custom,
                'bdist_egg': bdist_egg if 'bdist_egg' in sys.argv else bdist_egg_disabled,
                'sdist': pyvkfft_sdist,
                'install_lib': pyvkfft_install_lib
                },
      install_requires=install_requires,
      extras_require={'doc': ['sphinx', 'nbsphinx', 'nbsphinx-link', 'sphinx-argparse',
                              'sphinx-rtd-theme']},
      test_suite="test",
      entry_points={'console_scripts': console_scripts},
      )



# -*- coding: utf-8 -*-

__authors__ = ["Vincent Favre-Nicolin (pyvkfft), Dmitrii Tolmachev (VkFFT)"]
__license__ = "MIT"
__date__ = "2025/07/25"
# Valid numbering includes 3.1, 3.1.0, 3.1.2, 3.1dev0, 3.1a0, 3.1b0, 3.1.2.post1,...
__version__ = "2025.1.1"

import os


def vkfft_version():
    """
    Get VkFFT version
    :return: version as X.Y.Z
    """
    # We import here as otherwise it would mess with setup.py which reads __version__
    # while the opencl library has not yet been compiled.
    try:
        from .opencl import vkfft_version as cl_vkfft_version
        return cl_vkfft_version()
    except ImportError:
        # On some platforms (e.g. pp64le) opencl may not be available while cuda is
        try:
            from .cuda import vkfft_version as cu_vkfft_version
            return cu_vkfft_version()
        except ImportError:
            raise ImportError("Neither cuda or opencl vkfft_version could be imported")


def git_version():
    """
    Get the full pyvkfft version name with git hash, e.g. "2020.1-65-g958b7254-dirty"
    Only works if the current directory is part of the git repository, or
    after installation when the placeholder string has been replaced.
    :return: the pyvkfft git version, or "unknown"
    """
    # in distributed & installed versions this is replaced by a string
    __git_version_static__ = "2025.1.1"
    if "placeholder" not in __git_version_static__:
        return __git_version_static__
    from subprocess import Popen, PIPE
    try:
        p = Popen(['git', 'describe', '--tags', '--dirty', '--always'],
                  stdout=PIPE, stderr=PIPE)
        return p.stdout.readlines()[0].strip().decode("UTF-8")
    except:
        return "unknown"


def vkfft_git_version():
    """
    Get the full vkfft version name with git hash, e.g. "2020.1-65-g958b7254-dirty"
    Only works if the current directory is part of the git repository, or
    after installation when the placeholder string has been replaced.
    :return: the vkfft git version, or "unknown"
    """
    from subprocess import Popen, PIPE
    # in distributed & installed versions this is replaced by a string
    __git_version_static__ = "v1.3.4"
    if "placeholder" not in __git_version_static__:
        return __git_version_static__
    try:
        p = Popen(['git', 'describe', '--tags', '--dirty', '--always'],
                  cwd=os.path.join(os.getcwd(), 'src', 'VkFFT'), stdout=PIPE, stderr=PIPE)
        return p.stdout.readlines()[0].strip().decode("UTF-8")
    except:
        return "unknown"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join('..', '..')))
from pyvkfft.version import __authors__, __version__

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pyvkfft'
copyright = '2023- Vincent Favre-Nicolin ; ESRF-The European Synchrotron'
author = ', '.join(__authors__)
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.githubpages',
    'sphinx.ext.todo',
    'sphinx.ext.napoleon',
    'nbsphinx',
    'nbsphinx_link',
    'sphinxarg.ext'
]

templates_path = ['_templates']
exclude_patterns = []

nbsphinx_execute = 'never'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# Do not transform "--" into "-"
smartquotes = False